/*
 * A program that calculates the free energy of a strand of nucleic acids.
 * This strand of nucleic acids can be composed of either DNA or RNA.
 *
 * (c) 2008  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "efn2.h"

///////////////////////////////////////////////////////////////////////////////
// Constructor
///////////////////////////////////////////////////////////////////////////////
efn2Interface::efn2Interface() {

  // Initialize boolean flags.
  isRNA = true;
  writeTherm = false;
  stdPrint = false;

  // Initialize file names.
  ctFile = 0;
  outFile = 0;

  // Initialize temperature.
  temperature = 310.15;
}

///////////////////////////////////////////////////////////////////////////////
// Parse command line arguments.
///////////////////////////////////////////////////////////////////////////////
bool efn2Interface::parse( int argc, char** argv ) {

  // Initialize arrays of flags without and with parameters.
  const char* flagsNoParams[] = {
    "-d", "-D", "--DNA",
    "-p", "-P", "--print",
    "-w", "-w", "--writedetails"
  };

  const char* flagsParams[] = { "-t", "-T", "--temperature" };

  // Initialize usage string.
  const char* usageString = "USAGE: efn2 <ct file> <energy file> [options]";

  // Initialize required parameters string.
  string ctHelp = "\
<ct file>\n\
     The name of a file containing structure CT data.\
";

  string outHelp = "\
<output file>\n\
     The energy file to which output is written.\n\
     This output file can be written in one of two forms:\n\
     1. Simple list\n\
        Lists free energy for each structure, lowest first.\n\
     2. Thermodynamic details\n\
        Writes details of every substructure in each structure, and the\n\
        corresponding free energy of each.\
";

  string paramString = ctHelp + "\n\n" + outHelp;

  // Initialize flags without parameters string.
  string dnaHelp = "\
-d, -D, --DNA\n\
     Specify that the sequence if DNA, and DNA parameters are to be used.\n\
     Default is to use RNA parameters.\
";

  string helpLabel = "\
-h, -H, --help\n\
     Display the usage details message.\
";

  string printHelp = "\
-p, -P, --print\n\
     Print the output file to standard output.\n\
     This won't override default behavior of writing to a file.\n\
     Thermodynamic files (if written) are not piped.\
";

  string detailHelp = "\
-w, -W, --writedetails\n\
     Write a thermodynamic details file.\n\
     The thermodynamic details file replaces a standard output (list) file.\
";

  string flagsNoParamsString = dnaHelp + "\n\n" + helpLabel + "\n\n" +
    printHelp + "\n\n" + detailHelp;

  // Initialize the flags with parameters string.
  string tempHelp = "\
-t, -T, --temperature\n\
     Specify the temperature at which calculation takes place in Kelvin.\n\
     Default is 310.15 K, which is 37 C.\
";

  string flagsParamsString = tempHelp;

  // Initialize parser and check the command line.
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   9, flagsNoParams,
						   3, flagsParams );
  parser->setUsageStrings( usageString, paramString.c_str(),
                           flagsNoParamsString.c_str(),
			   flagsParamsString.c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing.
  if( noError ) {

    // Get required parameters.
    ctFile = parser->getParameter( 1 );
    outFile = parser->getParameter( 2 );

    // Set non-parameterized flags
    isRNA = !parser->containsInGroup( "-d -D --DNA" );
    stdPrint = parser->containsInGroup( "-p -P --print" );
    writeTherm = parser->containsInGroup( "-w -W --writedetails" );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Temperature option
    flagSet = parser->setOption( temperature, "-t -T --temperature",
                                 "Invalid temperature given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }
  
  // Delete the parser and return the error state
  delete parser;
  return noError;

}

///////////////////////////////////////////////////////////////////////////////
// Run efn2 calculations; print to standard output and/or write a thermodynamic
// details file as necessary.
///////////////////////////////////////////////////////////////////////////////
void efn2Interface::run() {

  // Create a variable to handle errors.
  int error = 0;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 1 (ct file).
   * isRNA identifies whether the strand is RNA (true) or DNA (false).
   *
   * After construction, check the error status of the strand using the
   * isErrorStatus function.
   * isErrorStatus will return 0 if there is no error.
   * Throughout, the calculation proceeds as long as error = 0.
   */
  cout << "Initializing nucleic acids..." << flush;
  RNA* strand = new RNA( ctFile, 1, isRNA );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Set the temperature using the SetTemperature method.
   */
  if( error == 0 ) {
    int tempError = strand->SetTemperature( temperature );
    error = checker->isErrorStatus( tempError );
  }

  /*
   * If the user wants a thermodynamic details file, write that alone because
   * it is simply a more detailed version of the default output file.
   */
  if( ( error == 0 ) && ( writeTherm ) ) {

    // Show a message saying that the details file is being written.
    cout << "Writing thermodynamic details file..." << flush;

    // Write the thermodynamic details file and check for errors.
    int thermError = strand->WriteThermodynamicDetails( outFile );
    error = checker->isErrorStatus( thermError );
  }

  /*
   * Otherwise, if the user wants a simple output file, write it, and pipe to
   * standard output if the user requests it.
   */
  else if( ( error == 0 ) && ( !writeTherm ) ) {

    // Show a message saying that the simple output file is being written.
    cout << "Writing free energy list file..." << flush;

    // Fetch the number of structures that are in the ct file.
    structures = strand->GetStructureNumber();

    // For each structure, calculate its energy. If no error occurred in this
    // calculation, put the energy value in the energies vector. If an error
    // did occur, set the error flag and stop calculations.
    for( int i = 1; i <= structures; i++ ) {
      double energy = strand->CalculateFreeEnergy( i );
      error = checker->isErrorStatus();

      if( error == 0 ) { energies.push_back( energy ); }
      else break;
    }

    // If all free energies were calculated correctly, write the output file.
    if( error == 0 ) {
      ofstream out( outFile );
      for( int i = 1; i <= structures; i++ ) {
	out << "Structure: " << i << "   Energy = " << fixed
	    << setprecision( 1 ) << energies.at( i - 1 ) << endl;
      }
    }
  }

  /*
   * If the free energy calculations, whichever type, were finished correctly,
   * print a message that says so.
   */
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * If the output is a simple list file and should be piped to standard output
   * then pipe it.
   */
  if( ( error == 0 ) && ( !writeTherm ) && ( stdPrint ) ) {

    // Write a header for the outputted file.
    cout << endl << "Generated output file: " << outFile << endl << endl;

    for( int i = 1; i <= structures; i++ ) {
      cout << "Structure: " << i << "   Energy = " << fixed
	   << setprecision( 1 ) << energies.at( i - 1 ) << endl;
    }

    // Add a blank line after the file is printed, for neatness.
    cout << endl;
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  if( error == 0 ) { cout << "efn2 complete." << endl; }
  else { cerr << "efn2 complete with errors." << endl; }
}

///////////////////////////////////////////////////////////////////////////////
// Main method to run the program.
///////////////////////////////////////////////////////////////////////////////
int main( int argc, char* argv[] ) {

  efn2Interface* myEfn2 = new efn2Interface();
  bool parseable = myEfn2->parse( argc, argv );
  if( parseable == true ) { myEfn2->run(); }
  delete myEfn2;
  return 0;

}
