/*
 * An implementation file for a program that finds pseudoknots in a strand of
 * nucleic acids.
 * This strand of nucleic acids can be composed of either DNA or RNA.
 *
 * (c) 2010  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "ProbKnot_Interface.h"

///////////////////////////////////////////////////////////////////////////////
// Constructor
///////////////////////////////////////////////////////////////////////////////
ProbKnot::ProbKnot() {

  // Set the input and output file names to null pointers.
  inFile = 0;
  ctFile = 0;

  // Set the default values of the additional parameters.
  isRNA = true;
  isSequence = false;

  iterations = 1;
  minHelixLength = 3;

}

///////////////////////////////////////////////////////////////////////////////
// Parse the command line.
///////////////////////////////////////////////////////////////////////////////
bool ProbKnot::parse( int argc, char** argv ) {

  // Initialize arrays of flags without and with parameters
  const char* flagsNoParams[] = {
    "-d", "-D", "--DNA",
    "--sequence"
  };

  const char* flagsParams[] = {
    "-i", "-I", "--iterations",
    "-m", "-M", "--minimum"
  };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString =
    "USAGE: ProbKnot <input file> <ct file> [options]";

  stringstream use1;
  use1 << "<input file>" << endl
       << "     Specifies an input file."
       << endl
       << "     Can be either a partition function save file (holds "
       <<"probability data) or a"
       << endl
       << "     sequence file (holds raw sequence)."
       << endl << endl
       << "<ct file>" << endl
       << "     Specifies ct file to which structures are written.";

  stringstream use2;
  use2 << "-d, -D, --DNA" << endl
       << "     Specifies the input file as holding DNA information." << endl
       << "     This flag only matters if the input file is a sequence file."
       << endl << endl
       << "-h, -H, --help" << endl
       << "     Displays this usage details message."
       << endl << endl
       << "--sequence" << endl
       << "     Specifies that the input file is a sequence file.";

  stringstream use3;
  use3 << "-i, -I, --iterations" << endl
       << "     Specifies number of iterations the calculation will undergo."
       << endl
       << "     Default is 1 iteration."
       << endl << endl
       << "-m, -M, --minimum" << endl
       << "     Specifies the minimum length accepted for a helix." << endl
       << "     Default is 3 base pairs.";

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
                                                   4, flagsNoParams,
                                                   6, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
                           use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {

    // Get the required parameters.
    inFile = parser->getParameter( 1 );
    ctFile = parser->getParameter( 2 );

    // Set non-parameterized flags. The RNA/DNA flag is set only if a sequence
    // flag is present.
    isSequence = parser->contains( "--sequence" );
    if( isSequence ) { isRNA = !parser->containsInGroup( "-d -D --DNA" ); }

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Iterations option
    flagSet =
      parser->setOption( iterations, "-i -I --iterations",
			 "Invalid number of iterations given.", "> 0" );
    if( flagSet == false ) { noError = false; }

    // Minimum helix length option
    flagSet =
      parser->setOption( minHelixLength, "-m -M --minimum",
			 "Invalid minimum helix length given.", "> 0" );
    if( flagSet == false ) { noError = false; }

  }

  // Delete the parser and return the error state
  delete parser;
  return noError;

}

///////////////////////////////////////////////////////////////////////////////
// Run ProbKnot calculations.
///////////////////////////////////////////////////////////////////////////////
void ProbKnot::run() {

  // Create a variable that keeps track of the error state, and set it to a
  // default of 0 (no error).
  int error = 0;

  /*
   * Use the constructor for RNA that specifies a filename.
   * If the input file is a pfs file, specify type = 3 (pfs file).
   * If the input file is a sequence file, specify type = 2 (sequence file).
   *
   * After construction of the strand, create the error checker which monitors
   * the strand for errors.
   * Throughout, the calculation proceeds as long as error = 0.
   */

  // Show a message that the RNA strand is being initialized.
  cout << "Initializing RNA strand..." << flush;

  // Set the type for the RNA strand, then create the strand itself.
  int type = ( !isSequence ) ? 3 : 2;
  RNA* strand = new RNA( inFile, type, isRNA );

  // Create the strand error checker.
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );

  // Check the error status and print a message finishing this section if no
  // error occurred.
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * If the input file is a sequence file, calculate the partition function
   * first before calculating pseudoknots.
   */
  if( error == 0 && isSequence ) {

    // Show a message that the partition function is being calculated.
    cout << "Calculating partition function..." << flush;

    // Run the partition function.
    error = strand->PartitionFunction();

    // Check the error status and print a message finishing this section if no
    // error occurred.
    if( !checker->isErrorStatus( error ) ) { cout << "done. " << endl; }
  }

  /*
   * Calculate pseudoknots in the strand using the ProbKnot method.
   * After calculation is done, check error status of strand as above.
   */
  if( error == 0 ) {

    // Show a message that pseudoknots are being calculated.
    cout << "Calculating pseudoknots..." << flush;

    // Run calculation of pseudoknots.
    error = strand->ProbKnot( iterations, minHelixLength );

    // Check the error status and print a message finishing this section if no
    // error occurred.
    if( !checker->isErrorStatus( error ) ) { cout << "done. " << endl; }
  }

  /*
   * Write a CT file for the strand.
   * Use method WriteCt to write the file.
   */
  if( error == 0 ) {
    // Show a message that a CT file is being written.
    cout << "Writing output ct file..." << flush;

    // Run writing of the CT file.
    error = strand->WriteCt( ctFile );

    // Check the error status and print a message finishing this section if no
    // error occurred.
    if( !checker->isErrorStatus( error ) ) { cout << "done." << endl; }
  }

  /*
   * Clean up the dynamically allocated RNA strand and error checker, then
   * print confirmation of run finishing. The confirmation message varies
   * depending on whether an error was encountered at any point.
   */

  // Delete the strand and error checker.
  delete strand;
  delete checker;

  // Print out the confirmation of run finishing.
  if( !error ) { cout << "Calculation of pseudoknots complete." << endl; }
  else { cerr << "Calculation of pseudoknots complete with errors." << endl; }

}

///////////////////////////////////////////////////////////////////////////////
// Main method to run the program.
///////////////////////////////////////////////////////////////////////////////
int main( int argc, char* argv[] ) {

  ProbKnot* knot = new ProbKnot();
  bool parseable = knot->parse( argc, argv );
  if( parseable ) { knot->run(); }
  delete knot;

  return 0;

}
