/*
 * A program that generates structures based on levels of probable pairs.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "ProbablePair.h"

// Constructor
ProbablePair::ProbablePair() {
  input = 0;
  output = 0;
  threshold = 0.0;
}

// Parse the command line arguments
bool ProbablePair::parse( int argc, char** argv ) {
  // Initialize array of flags with parameters
  const char* flagsParams[] = { "-t", "-T", "--temperature" };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString =
    "USAGE: ProbablePair <pfs file> <ct file> [options]";

  stringstream use1;
  use1 << "<pfs file>" << endl
       << "     Specifies the input pfs file to generate structures from."
       << endl << endl
       << "<ct file>" << endl
       << "     Specifies the output ct file where structures will be "
       << "written.";

  stringstream use2;
  use2 << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3;
  use3 << "-t, -T, --threshold" << endl
       << "     Specifies pairing threshold for pairs to be included." << endl
       << "     This should be expressed as a number >= 0.5 and <= 1.0."
       << endl
       << "     Default is 0, which signifies structures should be generated "
       << "at multiple" << endl
       << "     thresholds:" << endl
       << "          >= 0.99, >= 0.97, >= 0.95, >= 0.90, >= 0.80, >= 0.70, "
       << ">= 0.60, and" << endl
       << "          >= 0.50.";

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   0, NULL,
						   3, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
                           use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    input = parser->getParameter( 1 );
    output = parser->getParameter( 2 );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that action is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Temperature option
    flagSet = parser->setOption( threshold, "-t -T --threshold",
				 "Invalid threshold given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error status
  delete parser;
  return noError;
}

// Run probable structure generation
void ProbablePair::run() {
  int error = 0;

  // Create the RNA strand
  cout << "Initializing RNA strand..." << flush;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 3 (pfs file).
   *
   * After construction of the strand, create the error checker which monitors
   * the strand for errors
   * Throughout, the calculation proceeds as long as error = 0.
   */
  RNA* strand = new RNA( input, 3, true );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Predict probable pairs using the PredictProbablePairs method.
   * After prediction is done, check error status of strand as above.
   */
  if( error == 0 ) {
    // Predict structures
    cout << "Calculating probable structures..." << flush;
    error = strand->PredictProbablePairs( threshold );

    // Check the error status
    if( !checker->isErrorStatus( error ) ) { cout << "done." << endl; }
  }

  /*
   * Write a CT file for the strand
   * Use method WriteCt to write the file.
   *
   * After writing is complete, check the error status of the strand and print
   * a final message announcing the finish of the run.
   */
  if( error == 0 ) {
    // Write the CT file
    cout << "Writing output ct file..." << flush;
    int writeError = strand->WriteCt( output );

    // Check the error status
    error = checker->isErrorStatus( writeError );
    if( error == 0 ) { cout << "done." << endl; }
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  const char* calcType = "Calculation of probable structures";
  if( error == 0 ) { cout << calcType << " complete." << endl; }
  else { cerr << calcType << " complete with errors." << endl; } 
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  ProbablePair* myPair = new ProbablePair();
  bool parseable = myPair->parse( argc, argv );
  if( parseable == true ) { myPair->run(); }
  delete myPair;
  return 0;
}
