/*
 * A program that breaks pseudoknots to allow for easier analysis of a strand
 * of nucleic acids.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "RemovePseudoknots.h"

// Constructor
RemovePseudoknots::RemovePseudoknots() {
  input = 0;
  output = 0;
  temperature = 310.15;
  isRNA = true;
}

// Parse the command line arguments
bool RemovePseudoknots::parse( int argc, char** argv ) {
  // Initialize arrays of flags without and with parameters
  const char* flagsNoParams[] = { "-d", "-D", "--DNA" };
  const char* flagsParams[] = { "-t", "-T", "--temperature" };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString =
    "USAGE: RemovePseudoknots <input ct file> <output ct file> [options]";

  stringstream use1;
  use1 << "<input ct file>" << endl
       << "     Specifies the input ct file which contains pseudoknots."
       << endl << endl
       << "<output ct file>" << endl
       << "     Specifies the output ct file from which pseudoknots have been "
       << "removed.";

  stringstream use2;
  use2 << "-d, -D, --DNA" << endl
       << "     Specifies that the sequence is DNA, and DNA parameters are to "
       << "be used."
       << endl << endl
       << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3;
  use3 << "-t, -T, --temperature" << endl
       << "     Specifies the temperature at which calculation takes place in "
       << "Kelvin." << endl
       << "     Default is 310.15 K, which is 37 C.";

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   3, flagsNoParams,
						   3, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
                           use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    input = parser->getParameter( 1 );
    output = parser->getParameter( 2 );

    // Set non-parameterized DNA/RNA flag
    isRNA = !parser->containsInGroup( "-d -D --DNA" );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that action is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Temperature option
    flagSet = parser->setOption( temperature, "-t -T --temperature",
				 "Invalid temperature given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error status
  delete parser;
  return noError;
}

// Run pseudoknot breaking
void RemovePseudoknots::run() {
  int error = 0;

  // Create the RNA strand
  cout << "Initializing RNA strand..." << flush;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 1 (ct file).
   *
   * After construction of the strand, create the error checker which monitors
   * the strand for errors
   * Throughout, the calculation proceeds as long as error = 0.
   */
  RNA* strand = new RNA( input, 1, isRNA );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Set the temperature using the SetTemperature method.
   */
  if( error == 0 ) {
    error = strand->SetTemperature( temperature );
  }

  /*
   * Break pseudoknots using the BreakPseudoknot method.
   * After pseudoknot breakage is done, check error status of strand as above.
   */
  if( error == 0 ) {
    // Break pseudoknots
    cout << "Breaking pseudoknots..." << flush;
    error = strand->BreakPseudoknot();

    // Check the error status
    if( !checker->isErrorStatus( error ) ) { cout << "done." << endl; }
  }

  /*
   * Write a CT file for the strand cleaned of pseudoknots
   * Use method WriteCt to write the file.
   *
   * After writing is complete, check the error status of the strand and print
   * a final message announcing the finish of the run.
   */
  if( error == 0 ) {
    // Write the CT file
    cout << "Writing output ct file..." << flush;
    int writeError = strand->WriteCt( output );

    // Check the error status
    error = checker->isErrorStatus( writeError );
    if( error == 0 ) { cout << "done." << endl; }
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  if( error == 0 ) { cout << "RNA pseudoknot breakage complete." << endl; }
  else { cerr << "RNA pseudoknot breakage complete with errors." << endl; } 
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  RemovePseudoknots* myPseudoknot = new RemovePseudoknots();
  bool parseable = myPseudoknot->parse( argc, argv );
  if( parseable == true ) { myPseudoknot->run(); }
  delete myPseudoknot;
  return 0;
}
