/*
 * A program that folds a strand of nucleic acids.
 * This strand of nucleic acids can be composed of either DNA or RNA.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "Fold.h"

// Constructor
Fold::Fold() {
  isRNA = true;

  seqFile = 0;
  ctFile = 0;
  saveFile = "";

  constraintFile = 0;
  SHAPEFile = 0;

  singleOffsetFile = 0;
  doubleOffsetFile = 0;

  percent = 10;
  numStructures = 20;
  windowSize = -1;

  maxDistance = -1;
  maxLoop = 30;

  slope = 2.6;
  slopeSingle = 0;

  intercept = -0.8;
  interceptSingle = 0;

  temperature = 310.15;
}

// Parse the command line arguments
bool Fold::parse( int argc, char** argv ) {
  // Initialize arrays of flags without and with parameters
  const char* flagsNoParams[] = { "-d", "-D", "--DNA" };

  const char* flagsParams[] = {
    "-c", "-C", "--constraint",
    "-dso", "-DSO", "--doubleOffset",
    "-l", "-L", "--loop",
    "-m", "-M", "--maximum",
    "-md", "-MD", "--maxdistance",
    "-p", "-P", "--percent",
    "-s", "-S", "--save",
    "-sh", "-SH", "--SHAPE",
    "-si", "-SI", "--SHAPEintercept",
    "-sm", "-SM", "--SHAPEslope",
    "-sso", "-SSO", "--singleOffset",
    "-t", "-T", "--temperature",
    "-usi", "-USI", "--unpairedSHAPEintercept",
    "-usm", "-USM", "--unpairedSHAPEslope",
    "-w", "-W", "--window"
  };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3 and use4: Flags with parameters streams
  // The flags with parameters streams are split, then recombined, because
  // they're too large for a single stringstream.
  const char* usageString = "USAGE: Fold <seq file> <ct file> [options]";

  stringstream use1;
  use1 << "<seq file>" << endl
       << "     Specifies the sequence file that is folded."
       << endl << endl
       << "<ct file>" << endl
       << "     Specifies the ct file to which the fold is written.";

  stringstream use2;
  use2 << "-d, -D, --DNA" << endl
       << "     Specifies that the sequence is DNA, and DNA parameters are to "
       << "be used."
       << endl << endl
       << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3, use4;
  use3 << "-c, -C, --constraint" << endl
       << "     Specifies a folding constraints file to be applied." << endl
       << "     Default is no folding constraints applied."
       << endl << endl
       << "-l, -L, --loop" << endl
       << "     Specifies a maximum internal/bulge loop size." << endl
       << "     Default is 30 unpaired nucleotides."
       << endl << endl
       << "-m, -M, --maximum" << endl
       << "     Specifies a maximum number of structures." << endl
       << "     Default is 20 structures."
       << endl << endl
       << "-md, -MD, --maxdistance" << endl
       << "     Specifies a maximum pairing distance." << endl
       << "     Default is no restriction on distance between pairs."
       << endl << endl
       << "-p, -P, --percent" << endl
       << "     Specifies a percent difference." << endl
       << "     Default is 10 percent (ie. 10, not 0.1)."
       << endl << endl
       << "-s, -S, --save" << endl
       << "     Specifies a save file, used for dot plots and refolding."
       << endl
       << "     Default is no save file name specified."
       << endl << endl
       << "-sh, -SH, --SHAPE" << endl
       << "     Specifies a SHAPE constraints file to be applied." << endl
       << "     Default is no SHAPE constraints applied." << endl
       << "     This constraint utilizes the SHAPE pseudoenergy constraints."
       << endl << endl
       << "-si, -SI, --SHAPEintercept" << endl
       << "     Specifies an intercept used with SHAPE constraints." << endl
       << "     Default is -0.8 kcal/mol."
       << endl << endl;
  use4 << "-sm, -SM, --SHAPEslope" << endl
       << "     Specifies a slope used with SHAPE constraints." << endl
       << "     Default is 2.6 kcal/mol."
       << endl << endl
       << "-t, -T, --temperature" << endl
       << "     Specifies the temperature at which folding takes place in "
       << "Kelvin." << endl
       << "     Default is 310.15 K, which is 37 degrees C."
       << endl << endl
       << "-w, -W, --window" << endl
       << "     Specifies a window size." << endl
       << "     Default is determined according to the length of the "
       << "sequence.";

  char* paramString = new char[use1.str().length()+1];
  strcpy( paramString, use1.str().c_str() );

  char* noOptionString = new char[use2.str().length()+1];
  strcpy( noOptionString, use2.str().c_str() );

  char* optionString = new char[use3.str().length() + use4.str().length()+1];
  strcpy( optionString, use3.str().c_str() );
  strcat( optionString, use4.str().c_str() );

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
                                                   3, flagsNoParams,
                                                   45, flagsParams );
  parser->setUsageStrings( usageString, paramString,
                           noOptionString, optionString );
  
  delete[] paramString;
  delete[] noOptionString;
  delete[] optionString;

  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    seqFile = parser->getParameter( 1 );
    ctFile = parser->getParameter( 2 );

    // Set non-parameterized RNA/DNA flag
    isRNA = !parser->containsInGroup( "-d -D --DNA" );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Constraint (folding) file option
    flagSet = parser->setOption( constraintFile, "-c -C --constraint",
				 "Folding constraints file not found.", true );
    if( flagSet == false ) { noError = false; }

    // Double strand offset file option
    flagSet =
      parser->setOption( doubleOffsetFile, "-dso -DSO --doubleOffset",
                         "Double strand offset file not found.", true );
    if( flagSet == false ) { noError = false; }

    // Maximum loop size option
    flagSet = parser->setOption( maxLoop, "-l -L --loop",
				 "Invalid maximum loop size given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  
    // Maximum number of structures option
    flagSet = parser->setOption( numStructures, "-m -M --maximum",
				 "Invalid maximum structures given.", "> 0" );
    if( flagSet == false ) { noError = false; }

    // Maximum pairing distance option
    flagSet = parser->setOption( maxDistance, "-md -MD --maxdistance",
				 "Invalid maximum distance given.", "> 0" );
    if( flagSet == false ) { noError = false; }

    // Percent difference option
    flagSet = parser->setOption( percent, "-p -P --percent",
				 "Invalid percent difference given.", ">= 0" );
    if( flagSet == false ) { noError = false; }

    // Save file option
    flagSet = parser->setOption( saveFile, "-s -S --save",
				 "Error creating save file name.", false );
    if( flagSet == false ) { noError = false; }

    // SHAPE constraint file option
    flagSet = parser->setOption( SHAPEFile, "-sh -SH --SHAPE",
				 "SHAPE constraint file not found.", true );
    if( flagSet == false ) { noError = false; }

    // SHAPE intercept option
    flagSet = parser->setOption( intercept, "-si -SI --SHAPEintercept",
                                 "Invalid SHAPE intercept given." );
    if( flagSet == false ) { noError = false; }

    // Single strand offset file option
    flagSet = parser->setOption( singleOffsetFile, "-sso -SSO --singleOffset",
                                 "Single strand offset file not found.",
				 true );
    if( flagSet == false ) { noError = false; }

    // SHAPE slope option
    flagSet = parser->setOption( slope, "-sm -SM --SHAPEslope",
                                 "Invalid SHAPE slope given." );
    if( flagSet == false ) { noError = false; }

    // Temperature option
    flagSet = parser->setOption( temperature, "-t -T --temperature",
				 "Invalid temperature given.", ">= 0" );
    if( flagSet == false ) { noError = false; }

    // Unpaired SHAPE intercept option
    flagSet =
      parser->setOption( interceptSingle, "-usi -USI --unpairedSHAPEintercept",
                         "Invalid unpaired SHAPE intercept given." );
    if( flagSet == false ) { noError = false; }

    // Unpaired SHAPE slope option
    flagSet = parser->setOption( slopeSingle, "-usm -USM --unpairedSHAPEslope",
                                 "Invalid unpaired SHAPE slope given." );
    if( flagSet == false ) { noError = false; }

    // Window size option
    flagSet = parser->setOption( windowSize, "-w -W --window",
				 "Invalid window size given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error state
  delete parser;
  return noError;
}

// Run the folding calculations
void Fold::run() {
  int error = 0;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 2 (sequence file).
   * isRNA identifies whether the strand is RNA (true) or DNA (false).
   *
   * After construction of the strand, construct the error checker
   * which monitors the strand for errors. 
   * Throughout, the calculation proceeds as long as error = 0.
   */
  cout << "Initializing nucleic acids..." << flush;
  RNA* strand = new RNA( seqFile, 2, isRNA );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Set the window size, based on the length of the sequence given as input.
   * Only do this, though, if the size hasn't been set on the command line.
   * Use method GetSequenceLength to get the length of the sequence.
   * The window sizes in relation to the length are hardcoded values.
   */
  if( windowSize == -1 && error == 0 ) {
    int length = strand->GetSequenceLength();
    windowSize = ( length > 1200 ) ? 20 :
      ( length > 800 ) ? 15 :
      ( length > 500 ) ? 11 :
      ( length > 300 ) ? 7 :
      ( length > 120 ) ? 5 :
      ( length > 50 ) ? 3 : 2;
  }

  /*
   * Set the temperature using the SetTemperature method.
   */
  if( error == 0 ) {
    int tempError = strand->SetTemperature( temperature );
    error = checker->isErrorStatus( tempError );
  }

  /*
   * Set maximum pairing distance using the ForceMaximumPairingDistance method.
   */
  if( error == 0 && maxDistance != -1 ) {
    int distError = strand->ForceMaximumPairingDistance( maxDistance );
    error = checker->isErrorStatus( distError );
  }

  /*
   * Add folding constraints and SHAPE constraints if a file has been given for
   * their inclusion.
   * For folding constraints, use the ReadConstraints method.
   * For SHAPE constraints, use the ReadSHAPE method.
   * For single strand offset, use the ReadSSO method.
   * For double strand offset, use the ReadDSO method.
   * After each constraint type, check the error status of the strand as above.
   */

  bool applyConstraints =
    ( constraintFile != 0 ) ||
    ( SHAPEFile != 0 ) ||
    ( singleOffsetFile != 0 ) ||
    ( doubleOffsetFile != 0 );

  if( error == 0 && applyConstraints ) {
    cout << "Applying constraints..." << flush;
    int constraintError = 0;

    // Read folding constraints
    if( constraintFile != 0 ) {
      constraintError = strand->ReadConstraints( constraintFile );
      error = checker->isErrorStatus( constraintError );
    }
    
    // Read SHAPE constraints
    if( error == 0 && SHAPEFile != 0 ) {
      constraintError = strand->ReadSHAPE( SHAPEFile, slope, intercept,
                                           slopeSingle, interceptSingle );
      error = checker->isErrorStatus( constraintError );
    }

    // Read single strand offset
    if( error == 0 && singleOffsetFile != 0 ) {
      constraintError = strand->ReadSSO( singleOffsetFile );
      error = checker->isErrorStatus( constraintError );
    }

    // Read double strand offset
    if( error == 0 && doubleOffsetFile != 0 ) {
      constraintError = strand->ReadDSO( doubleOffsetFile );
      error = checker->isErrorStatus( constraintError );
    }

    if( error == 0 ) { cout << "done." << endl; }
  }

  /*
   * Fold the single strand using the FoldSingleStrand method.
   * During calculation, monitor progress with the TProgressDialog class and
   * the Start/StopProgress methods of the RNA class.
   * After the strand is folded, check the error status of the strand as above.
   */
  if( error == 0 ) {
    cout << "Folding single strand..." << endl;

    // Create the progress monitor
    TProgressDialog* progress = new TProgressDialog();
    strand->SetProgress( *progress );

    // Fold the strand
    error = strand->FoldSingleStrand( percent, numStructures, windowSize,
                                      saveFile, maxLoop );
    error = checker->isErrorStatus( error );

    // Delete the progress monitor
    strand->StopProgress();
    delete progress;

    if( error == 0 ) { cout << "done." << endl; }
  }

  /*
   * Write a CT file from this fold.
   * Use method WriteCt to write the file.
   *
   * After writing is complete, check the error status of the strand and print
   * a final message announcing the finish of the run.
   */
  if( error == 0 ) {
    // Write the CT file
    cout << "Writing output ct file..." << flush;
    int writeError = strand->WriteCt( ctFile );

    // Check the error status
    error = checker->isErrorStatus( writeError );
    if( error == 0 ) { cout << "done." << endl; }
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  if( error == 0 ) { cout << "Single strand folding complete." << endl; }
  else { cerr << "Single strand folding complete with errors." << endl; } 
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  Fold* myFold = new Fold();
  bool parseable = myFold->parse( argc, argv );
  if( parseable == true ) { myFold->run(); }
  delete myFold;
  return 0;
}
