/*
 * A program that calculates thermodynamic parameters from a set of
 * oligonucleotides.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "oligoscreen.h"

// Constructor
oligoscreen::oligoscreen() {
  listFile = 0;
  reportFile = 0;

  temperature = 310.15;
  isRNA = true;
}

// Parse the command line arguments
bool oligoscreen::parse( int argc, char** argv ) {
  // Initialize arrays of flags without and with parameters
  const char* flagsNoParams[] = { "-d", "-D", "--DNA" };
  const char* flagsParams[] = { "-t", "-T", "--temperature" };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString =
    "USAGE: oligoscreen <list file> <report file> [options]";

  stringstream use1;
  use1 << "<list file>" << endl
       << "     Specifies the file which holds the oligo list."
       << endl << endl
       << "<report file>" << endl
       << "     Specifies a report file to which thermodynamics are written.";

  stringstream use2;
  use2 << "-d, -D, --DNA" << endl
       << "     Specifies that the oligomers are DNA, and DNA chemistry is to "
       << "be used."
       << endl << endl
       << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3;
  use3 << "-t, -T, --temperature" << endl
       << "     Specifies the temperature at which screening takes place in "
       << "Kelvin." << endl
       << "     Default is 310.15 K, which is 37 degrees C.";

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   3, flagsNoParams,
						   3, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
			   use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    listFile = parser->getParameter( 1 );
    reportFile = parser->getParameter( 2 );

    // Set non-parameterized chemistry flag
    isRNA = !parser->containsInGroup( "-d -D --DNA" );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Temperature option
    flagSet = parser->setOption( temperature, "-t -T --temperature",
                                 "Invalid temperature given." ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error state
  delete parser;
  return noError;
}

// Run the OligoScreen calculations
void oligoscreen::run() {
  // Create objects
  Oligowalk_object* screen = new Oligowalk_object( isRNA );
  ErrorChecker<Oligowalk_object>* checker =
    new ErrorChecker<Oligowalk_object>( screen );

  // Create variables
  int code = 0;
  bool error = false;

  // If OligoScreen object was created successfully, set temperature
  if( !( error = checker->isErrorStatus() ) ) {
    code = screen->SetTemperature( temperature );
    error = checker->isErrorStatus( code );
  }

  // If no errors have been encountered, run OligoScreen
  if( !error ) {
    code = screen->OligoScreen( listFile, reportFile );
    checker->isErrorStatus( code );
  }

  // Delete the checker and strand
  delete checker;
  delete screen;

  // Print confirmation of run finishing
  if( !error ) { cout << "OligoScreen complete." << endl; }
  else { cerr << "OligoScreen complete with errors." << endl; }
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  oligoscreen* myOligoScreen = new oligoscreen();
  bool parseable = myOligoScreen->parse( argc, argv );
  if( parseable == true ) { myOligoScreen->run(); }
  delete myOligoScreen;
  return 0;
}
