/*
 * A program that calculates the partition function for a strand of nucleic
 * acids.
 * This strand of nucleic acids can be composed of either DNA or RNA.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "partition.h"

// Constructor
partitionInterface::partitionInterface() {
  seqFile = 0;
  pfsFile = 0;

  constraintFile = 0;
  SHAPEFile = 0;

  slope = 2.6;
  intercept = -0.8;
  distance = -1;

  temperature = 310.15;
  isRNA = true;
}

// Parse the command line arguments
bool partitionInterface::parse( int argc, char** argv ) {
  // Initialize arrays of flags without and with parameters
  const char* flagsNoParams[] = { "-d", "-D", "--DNA" };

  const char* flagsParams[] = {
    "-c", "-C", "--constraint",
    "-md", "-MD", "maxdistance",
    "-sh", "-SH", "--SHAPE",
    "-si", "-SI", "--SHAPEintercept",
    "-sm", "-SM", "--SHAPEslope",
    "-t", "-T", "--temperature"
  };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString = "USAGE: partition <seq file> <pfs file> [options]";

  stringstream use1;
  use1 << "<seq file>" << endl
       << "     Specifies the sequence file on which the partition function "
       << "is calculated."
       << endl << endl
       << "<pfs file>" << endl
       << "     Specifies the save file to which the results of the partition "
       << "calculation" << endl
       << "     are written.";

  stringstream use2;
  use2 << "-d, -D, --DNA" << endl
       << "     Specifies that the sequence is DNA, and DNA parameters are to "
       << "be used."
       << endl << endl
       << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3;
  use3 << "-c, -C, --constraint" << endl
       << "     Specifies a folding constraints file to be applied." << endl
       << "     Default is no folding constraints applied."
       << endl << endl
       << "-md, -MD, --maxdistance" << endl
       << "     Specifies a maximum pairing distance." << endl
       << "     Default is no restriction on distance between pairs."
       << endl << endl
       << "-sh, -SH, --SHAPE" << endl
       << "     Specifies a SHAPE constraints file to be applied." << endl
       << "     Default is no SHAPE constraints applied." << endl
       << "     This constraint utilizes the SHAPE pseudoenergy constraints."
       << endl << endl
       << "-si, -SI, --SHAPEintercept" << endl
       << "     Specifies an intercept used with SHAPE constraints." << endl
       << "     Default is -0.8 kcal/mol."
       << endl << endl
       << "-sm, -SM, --SHAPEslope" << endl
       << "     Specifies a slope used with SHAPE constraints." << endl
       << "     Default is 2.6 kcal/mol."
       << endl << endl
       << "-t, -T, --temperature" << endl
       << "     Specifies the temperature at which folding takes place in "
       << "Kelvin." << endl
       << "     Default is 310.15 K, which is 37 degrees C.";

  // Initialize parser and check command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   3, flagsNoParams,
						   18, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
                           use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    seqFile = parser->getParameter( 1 );
    pfsFile = parser->getParameter( 2 );

    // Set non-parameterized RNA/DNA flag
    isRNA = !parser->containsInGroup( "-d -D --DNA" );

    // Set parameterized options
    // If one of these optionss is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all error are shown if there are more than one.
    bool flagSet = true;

    // Constraint (folding) file option
    flagSet = parser->setOption( constraintFile, "-c -C --constraint",
                                 "Folding constraint file not found.", true );
    if( flagSet == false ) { noError = false; }

    // Maximum pairing distance option
    flagSet =
      parser->setOption( distance, "-m -M --maxdistance",
                         "Invalid maximum pairing distance given.", "> 0" );
    if( flagSet == false ) { noError = false; }

    // SHAPE constraint file option
    flagSet = parser->setOption( SHAPEFile, "-sh -SH --SHAPE",
                                 "SHAPE constraint file not found.", true );
    if( flagSet == false ) { noError = false; }

    // SHAPE intercept option
    flagSet = parser->setOption( intercept, "-si -SI --SHAPEintercept",
                                 "Invalid SHAPE intercept given." );
    if( flagSet == false ) { noError = false; }

    // SHAPE slope option
    flagSet = parser->setOption( slope, "-sm -SM -SHAPEslope",
                                 "Invalid SHAPE slope given." );
    if( flagSet == false ) { noError = false; }

    // Temperature option
    flagSet = parser->setOption( temperature, "-t -T --temperature",
                                 "Invalid temperature given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error state
  delete parser;
  return noError;
}

// Run the partition function calculation
void partitionInterface::run() {
  int error = 0;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 2 (sequence file).
   * isRNA identifies whether the strand is RNA (true) or DNA (false).
   *
   * After construction, check the error status of the strand using the
   * isErrorStatus function.
   * isErrorStatus will return 0 if there is no error.
   * Throughout, the calculation proceeds as long as error = 0.
   */
  cout << "Initializing nucleic acids..." << flush;
  RNA* strand = new RNA( seqFile, 2, isRNA );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Set the temperature using the SetTemperature method.
   */
  if( error == 0 ) {
    int tempError = strand->SetTemperature( temperature );
    error = checker->isErrorStatus( tempError );
  }

  /*
   * Add folding constraints and SHAPE constraints if a file has been given for
   * their inclusion.
   * For folding constraints, use the ReadConstraints method.
   * For SHAPE constraints, use the ReadSHAPE method.
   * After each constraint type, check the error status of the strand as above.
   */
  bool applyConstraints =
    constraintFile != 0 ||
    SHAPEFile != 0;

  if( error == 0 && applyConstraints ) {
    cout << "Applying constraints..." << flush;
    int constraintError = 0;

    // Read folding constraints
    if( constraintFile != 0 ) {
      constraintError = strand->ReadConstraints( constraintFile );
      error = checker->isErrorStatus( constraintError );
    }
    
    // Read SHAPE constraints
    if( error == 0 && SHAPEFile != 0 ) {
      constraintError = strand->ReadSHAPE( SHAPEFile, slope, intercept );
      error = checker->isErrorStatus( constraintError );
    }

    if( error == 0 ) { cout << "done." << endl; }
  }

  /*
   * If a maximum pairing distance has been specified, set it using the
   * ForceMaximumPairingDistance method.
   * Provide a means to report calculation progress with the TProgressDialog
   * class and the Start/StopProgress methods of the RNA class.
   * Calculate the partition function using the PartitionFunction method.
   * After calculation is done, check the error status of the strand as above.
   */
  if( error == 0 ) {
    cout << "Calculating partition function..." << endl;

    // Set the maximum pairing distance if necessary
    if( distance != -1 ) {
      int distError = strand->ForceMaximumPairingDistance( distance );
      error = checker->isErrorStatus( distError );
    }

    // Prepare for calculating partition function, then calculate it while
    // using a TProgressDialog to monitor progress.
    if( error == 0 ) {
      TProgressDialog* progress = new TProgressDialog(); 
      strand->SetProgress( *progress );

      int partError = strand->PartitionFunction( pfsFile );
      error = checker->isErrorStatus( partError );

      strand->StopProgress();
      delete progress;

      if( error == 0 ) { cout << "done." << endl; }
    }
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  if( error == 0 ) { cout << "Single strand partitioning complete." << endl; }
  else { cerr << "Single strand partitioning complete with errors." << endl; } 
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  partitionInterface* myPartition = new partitionInterface();
  bool parseable = myPartition->parse( argc, argv );
  if( parseable == true ) { myPartition->run(); }
  delete myPartition;
  return 0;
}
