/*
 * A program that folds a strand of nucleic acids from a previously folded save
 * file.
 * This strand of nucleic acids can be composed of either DNA or RNA.
 *
 * (c) 2009  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#include "refold.h"

// Constructor
refold::refold() {
  saveFile = 0;
  ctFile = 0;

  percent = 10;
  numStructures = 20;
  windowSize = -1;
}

// Parse the command line arguments
bool refold::parse( int argc, char** argv ) {
  // Initialize array of flags with parameters
  const char* flagsParams[] = {
    "-m", "-M", "--maximum",
    "-p", "-P", "--percent",
    "-w", "-W", "--window",
  };

  // Initialize usage strings
  // usageString: Usage string
  // use1: Parameters string stream
  // use2: Flags without parameters stream
  // use3: Flags with parameters stream
  const char* usageString = "USAGE: refold <save file> <ct file> [options]";

  stringstream use1;
  use1 << "<save file>" << endl
       << "     Specifies the save file which holds a strand to be folded."
       << endl << endl
       << "<ct file>" << endl
       << "     Specifies the ct file to which the fold is written.";

  stringstream use2;
  use2 << "-h, -H, --help" << endl
       << "     Displays this usage details message.";

  stringstream use3;
  use3 << "-m, -M, --maximum" << endl
       << "     Specifies a maximum number of structures." << endl
       << "     Default is 20 structures."
       << endl << endl
       << "-p, -P, --percent" << endl
       << "     Specifies a percent difference." << endl
       << "     Default is 10 percent (ie. 10, not 0.1)."
       << endl << endl
       << "-w, -W, --window" << endl
       << "     Specifies a window size." << endl
       << "     Default is determined according to the length of the "
       << "sequence.";

  // Initialize parser and check the command line
  ParseCommandLine* parser = new ParseCommandLine( argc, argv, 2,
						   0, NULL,
						   9, flagsParams );
  parser->setUsageStrings( usageString, use1.str().c_str(),
                           use2.str().c_str(), use3.str().c_str() );
  bool noError = parser->checkLine();

  // If command line structure is OK proceed with argument retrieval, parsing
  if( noError ) {
    saveFile = parser->getParameter( 1 );
    ctFile = parser->getParameter( 2 );

    // Set parameterized options
    // If one of these options is not found on the command line, the action to
    // set that option is skipped.
    // If one throws an error, an error message is shown and parsing continues,
    // so all errors are shown if there are more than one.
    bool flagSet = true;

    // Maximum number of structures option
    flagSet = parser->setOption( numStructures, "-m -M --maximum",
                                 "Invalid maximum structures given." "> 0" );
    if( flagSet == false ) { noError = false; }

    // Percent difference flag
    flagSet = parser->setOption( percent, "-p -P --percent",
                                 "Invalid percent difference given.", ">= 0" );
    if( flagSet == false ) { noError = false; }

    // Window size flag
    flagSet = parser->setOption( windowSize, "-w -W --window",
                                 "Invalid window size given.", ">= 0" );
    if( flagSet == false ) { noError = false; }
  }

  // Delete the parser and return the error state
  delete parser;
  return noError;
}

// Run the refolding calculations
void refold::run() {
  int error = 0;

  // Create the strand
  cout << "Initializing nucleic acids..." << flush;

  /*
   * Use the constructor for RNA that specifies a filename.
   * Specify type = 3 (folding save file).
   * The save file handles the type of nucleic acid being folded, so it can be
   * set as the default (RNA) in the constructor.
   *
   * After construction, check the error status of the strand using the
   * isErrorStatus function.
   * isErrorStatus will return 0 if there is no error.
   * Throughout, the calculation proceeds as long as error = 0.
   */
  RNA* strand = new RNA( saveFile, 4 );
  ErrorChecker<RNA>* checker = new ErrorChecker<RNA>( strand );
  error = checker->isErrorStatus();
  if( error == 0 ) { cout << "done." << endl; }

  /*
   * Set the window size, based on the length of the sequence given as input.
   * Only do this, though, if window size hasn't been set on the command line.
   * Use method GetSequenceLength to get the length of the sequence.
   * The window sizes in relation to the length are hardcoded values.
   */
  if( windowSize == -1 && error == 0 ) {
    int length = strand->GetSequenceLength();
    windowSize = ( length > 1200 ) ? 20 :
      ( length > 800 ) ? 15 :
      ( length > 500 ) ? 11 :
      ( length > 300 ) ? 7 :
      ( length > 120 ) ? 5 :
      ( length > 50 ) ? 3 : 2;
  }

  /*
   * Refold the regenerated strand using the ReFoldSingleStrand method.
   * After the strand is folded, check the error status of the strand as above.
   */
  if( error == 0 ) {
    // Refold the strand
    cout << "Refolding nucleic acids..." << flush;
    error = strand->ReFoldSingleStrand( percent, numStructures, windowSize );

    // Check the error status
    error = checker->isErrorStatus( error );
    if( error == 0 ) { cout << "done." << endl; }
  }

  /*
   * Write a CT file from this refolding.
   * Use method WriteCt to write the file.
   *
   * After writing is complete, check the error status of the strand and print
   * a final message announcing the finish of the run.
   */
  if( error == 0 ) {
    // Write the CT file
    cout << "Writing output ct file..." << flush;
    int writeError = strand->WriteCt( ctFile );

    // Check the error status
    error = checker->isErrorStatus( writeError );
    if( error == 0 ) { cout << "done." << endl; }
  }

  // Delete the strand and checker, then print confirmation of run finishing
  delete checker;
  delete strand;

  if( error == 0 ) { cout << "Refolding complete." << endl; }
  else { cerr << "Refolding complete with errors." << endl; } 
}

// Main method to run the program
int main( int argc, char* argv[] ) {
  refold* myFold = new refold();
  bool parseable = myFold->parse( argc, argv );
  if( parseable == true ) { myFold->run(); }
  delete myFold;
  return 0;
}
