/*
 * A header file for a class that handles an object written in Postscript.
 *
 * (c) 2010  Mathews Lab, University of Rochester Medical Center
 * Written by Jessica Reuter
 */

#ifndef POSTSCRIPT_OBJECT_H
#define POSTSCRIPT_OBJECT_H

#include <iomanip>
#include <limits>
#include <sstream>
#include <string>
#include <vector>

#include "../../RNA_class/RNA.h"
#include "../ErrorChecker.h"

using namespace std;

class Postscript_Object {
 public:
  // Public error checking method

  /*
   * Name:        isError
   * Description: Tells if this object has encountered an error.
   * Returns:
   *     True if an error has been encountered, false if not.
   */
  bool isError();

 protected:
  // Protected constructor, destructor, and trimming method

  /*
   * Name:        Constructor
   * Descripton:  Initalizes private variables.
   */
  Postscript_Object();

  /*
   * Name:        Destructor
   * Description: Deletes all dynamically allocated variables.
   */
  ~Postscript_Object();

  /*
   * Name:        escapeAndTrim
   * Description: Escape parentheses characters "(" and ")" to ensure a string
   *              is safe to print in Postscript. In this function, "trim" has
   *              two meanings. First, it removes excess whitespace from the
   *              beginning and end of the string. Second, if the string is
   *              long enough where it will flow off the page, the string is
   *              truncated.
   *              Mode TRUNC_BEGIN should be used for file paths; it truncates
   *              a string like the following example:
   *              /dir1/dir2/dir3/dir4/dir5/file.txt -> PATH/file.txt
   *              Mode TRUNC_END should be used for any other string type; it
   *              truncates like the following example:
                  This is a string to truncate -> This is a string...
   * Arguments:
   *     1. data
   *        The string to be escaped.
   *     2. mode
   *        An int that tells what kind of truncation to do.
   *        Default is 0/NO_TRUNC (no truncation).
   * Returns:
   *     A string which holds the string with proper escapes
   */
  string escapeAndTrim( string data, int node = NO_TRUNC );

 protected:
  // Protected boolean that tells if an error occurred at any time.
  bool error;

 public:
  // Public static constant declarations; all classes that use the Postscript
  // hierarchy can access them.

  // Declarations of strand types
  static const int CT_TYPE  = 1;
  static const int SEQ_TYPE = 2;
  static const int PFS_TYPE = 3;
  static const int SAV_TYPE = 4;

  // Declarations of truncation modes for strings
  static const int NO_TRUNC = 0;
  static const int TRUNC_BEGIN = 1;
  static const int TRUNC_END = 2;

  // Declaration of static string that serves as a tab for Postscript writing
  static const string tab;

  // Declarations of static constants which hold Postscript colors
  static const string BLACK;
  static const string BLUE;
  static const string BRIGHT_GREEN;
  static const string DARK_PINK;
  static const string DARK_YELLOW;
  static const string GRAY;
  static const string GREEN;
  static const string LIGHT_BLUE;
  static const string ORANGE;
  static const string PURPLE;
  static const string RED;
  static const string WHITE;
};

#endif /* POSTSCRIPT_OBJECT_H */
