/////////////////////////////////////////////////////////////////
// Contrafold.cc
/////////////////////////////////////////////////////////////////

// Include files

#ifdef MULTI
#include <mpi.h>
#endif

#include <cerrno>

#include "Utilities.h"
#include "OptimizationWrapper.h"

using namespace std;

// Constants

const double GAMMA_DEFAULT = 6;

#ifdef SANITY
const double REGULARIZATION_DEFAULT = 1;
#else
const double REGULARIZATION_DEFAULT = 0;
#endif

// Global variables

double gammar = GAMMA_DEFAULT;
double posterior_cutoff = 0;
bool toggle_check = false;
bool toggle_train = false;
bool toggle_verbose = false;
bool toggle_viterbi = false;
bool toggle_partition = false;
bool toggle_complementary_only = false;
bool toggle_use_constraints = false;
double regularization_parameter = REGULARIZATION_DEFAULT;

vector<string> filenames;
string params_filename = "";
string output_parens_filename = "";
string output_bpseq_filename = "";
string output_posteriors_filename = "";

// Function prototypes

void Usage();
void Version();
void ParseParameters (int argc, char **argv);

/////////////////////////////////////////////////////////////////
// main()
//
// Main program.
/////////////////////////////////////////////////////////////////

int main (int argc, char **argv){

  int id = 0, num_procs = 1;

#ifdef MULTI
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &id);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
#endif
  
  ParseParameters (argc, argv);

  Parameters base_params;
  if (params_filename != "") base_params.ReadFromFile (params_filename);
  if (params_filename == "" && !toggle_train) base_params.InitializeDefaults();
  
  if (id == 0){
    OptimizationWrapper wrapper (id, num_procs, base_params);
    Parameters initial_params;

    if (toggle_train){
      initial_params.InitializeRandomly();

      if (regularization_parameter >= 0){
	wrapper.Train (initial_params, filenames, regularization_parameter);
      } else {
	wrapper.CrossValidate (initial_params, filenames, -regularization_parameter);
      }
      initial_params.LoadValues (initial_params.GetValues() + base_params.GetValues()).
	WriteToFile ("optimize.params.final");
      
    } else {

      if (gammar < 0){

	string old_output_parens_filename = output_parens_filename;
	string old_output_bpseq_filename = output_bpseq_filename;
	string old_output_posteriors_filename = output_posteriors_filename;

	if (output_parens_filename != "" && system (("mkdir -p " + output_parens_filename).c_str())) 
	  Error ("Unable to create directory for writing predictions.");
	if (output_bpseq_filename != "" && system (("mkdir -p " + output_bpseq_filename).c_str())) 
	  Error ("Unable to create directory for writing predictions.");
	if (output_posteriors_filename != "" && system (("mkdir -p " + output_posteriors_filename).c_str())) 
	  Error ("Unable to create directory for writing predictions.");

        for (int k = -5; k <= 10; k++){
          gammar = pow (2.0, (double) k);

	  if (old_output_parens_filename != "")
	    output_parens_filename = 
	      SPrintF ("%s/%s.gamma=%lf", old_output_parens_filename.c_str(), old_output_parens_filename.c_str(), gammar);
	  if (old_output_bpseq_filename != "")
	    output_bpseq_filename = 
	      SPrintF ("%s/%s.gamma=%lf", old_output_bpseq_filename.c_str(), old_output_bpseq_filename.c_str(), gammar);
	  if (old_output_posteriors_filename != "")
	    output_posteriors_filename = 
	      SPrintF ("%s%s.gamma=%lf", old_output_posteriors_filename.c_str(), old_output_posteriors_filename.c_str(), gammar);

          wrapper.DoPredictions (initial_params, filenames, gammar);
        }

	output_parens_filename = old_output_parens_filename;
	output_bpseq_filename = old_output_bpseq_filename;
	output_posteriors_filename = old_output_posteriors_filename;
	
      } else {

        wrapper.DoPredictions (initial_params, filenames, gammar);

      }
    }
  } else {
    OptimizationWrapper wrapper (id, num_procs, base_params);
    
#ifdef MULTI
    wrapper.RunComputeNode();
#endif
  }
  
#ifdef MULTI
  MPI_Finalize();
#endif
}

/////////////////////////////////////////////////////////////////
// Usage()
//
// Display program usage.
/////////////////////////////////////////////////////////////////

void Usage(){
  cerr << endl
       << "Usage: contrafold [predict|train] [OPTION]... INFILE(s)" << endl 
       << endl
       << "       where [OPTION]...  is a list of zero or more optional arguments" << endl
       << "             INFILE(s)    is the name of the input BPSEQ, parenthesized, or FASTA file(s)" << endl 
       << endl
       << "Miscellaneous arguments:" << endl
       << "  --version               display program version information" << endl
       << "  --verbose               show detailed console output" << endl
       << "  --params FILENAME       use particular model parameters" << endl
       << endl
       << "Additional arguments for 'predict' mode:" << endl
       << "  --gamma GAMMA           set sensivity/specificity tradeoff parameter (default: GAMMA=" << GAMMA_DEFAULT << ")" << endl
       << "                            if GAMMA > 1, emphasize sensitivity" << endl
       << "                            if 0 <= GAMMA <= 1, emphasize specificity" << endl
       << "                            if GAMMA < 0, try tradeoff parameters of 2^k for k = -5,...,10" << endl
       << "  --viterbi               use Viterbi instead of posterior decoding" << endl
       << "  --complementary         report only AU, CG, or GU pairs" << endl
       << "  --constraints           use existing constraints (requires BPSEQ or parenthesized format input)" << endl
       << endl
       << "  --parens OUTFILEORDIR   write parenthesized output to file or directory" << endl
       << "  --bpseq OUTFILEORDIR    write BPSEQ output to file or directory" << endl
       << "  --posteriors CUTOFF OUTFILEORDIR" << endl
       << "                          write posterior pairing probabilities to file or directory" << endl
       << "  --partition             compute the partition function only" << endl
       << endl
       << "Additional arguments for training (many input files may be specified):" << endl
       << "  --check                 check only for valid parses" << endl
       << "  --regularize C          perform BFGS training" << endl
       << "                            if C >= 0, use L2 regularization with coefficient C" << endl
       << "                            if C < 0, use fraction -C of training data for holdout cross-validation" << endl
       << endl;
  exit(0);
}

/////////////////////////////////////////////////////////////////
// Version()
//
// Display program version.
/////////////////////////////////////////////////////////////////

void Version(){
  cerr << "CONTRAFold version 1.10 - RNA secondary structure prediction" << endl << endl
       << "Written by Chuong B. Do" << endl;
  exit(0);
}

/////////////////////////////////////////////////////////////////
// ParseParameters()
//
// Parse command line parameters.
/////////////////////////////////////////////////////////////////

void ParseParameters (int argc, char **argv){
  if (argc < 2) Usage();
  filenames.clear();

  // Check for prediction or training mode

  if (string(argv[1]) == "train"){
    toggle_train = true;
  } else {
    if (string (argv[1]) != "predict")
      Error ("CONTRAfold must be run in either 'predict' or 'train' mode.");
  }

  for (int argno = 2; argno < argc; argno++){
    if (argv[argno][0] == '-'){

      // Miscellaneous options

      if (!strcmp (argv[argno], "--version")){
        Version();
      } else if (!strcmp (argv[argno], "--verbose")){
        toggle_verbose = true;
      } else if (!strcmp (argv[argno], "--params")){
        if (argno == argc - 1) Error ("Must specify FILENAME after --params.");
        params_filename = argv[++argno];
      } else if (!strcmp (argv[argno], "--partition")){
	toggle_partition = true;
      }

      // Prediction options

      else if (!strcmp (argv[argno], "--gamma")){
        if (argno == argc - 1) Error ("Must specify trade-off parameter GAMMA after --gamma.");
        if (!ConvertToDouble (argv[++argno], gammar))
          Error ("Unable to parse value after --gamma.");
      } else if (!strcmp (argv[argno], "--viterbi")){
        toggle_viterbi = true;
      } else if (!strcmp (argv[argno], "--complementary")){
	toggle_complementary_only = true;
      } else if (!strcmp (argv[argno], "--constraints")){
	toggle_use_constraints = true;
      } else if (!strcmp (argv[argno], "--parens")){
        if (argno == argc - 1) Error ("Must specify output file or directory name after --parens.");
	output_parens_filename = argv[++argno];
      } else if (!strcmp (argv[argno], "--bpseq")){
        if (argno == argc - 1) Error ("Must specify output file or directory name after --bpseq.");
	output_bpseq_filename = argv[++argno];
      } else if (!strcmp (argv[argno], "--posteriors")){
        if (argno == argc - 1) Error ("Must specify posterior probability threshold CUTOFF after --posteriors.");
        if (!ConvertToDouble (argv[++argno], posterior_cutoff))
	  Error ("Unable to parse cutoff value after --posteriors.");
        if (argno == argc - 1) Error ("Must specify output file or directory for --posteriors.");
	output_posteriors_filename = argv[++argno];
      }
      
      // Training options

      else if (!strcmp (argv[argno], "--check")){
        toggle_check = true;
      } else if (!strcmp (argv[argno], "--regularize")){
        if (argno == argc - 1) Error ("Must specify regularization parameter C after --regularize.");
        if (!ConvertToDouble (argv[++argno], regularization_parameter))
          Error ("Unable to parse regularization parameter after --regularize.");
        if (regularization_parameter <= -1)
          Error ("Invalid regularization parameter.  Value must be greater than -1.");
      } else {
        Error (("Unknown option \"" + string(argv[argno]) + "\" specified.  Run program without any arguments to see command-line options.").c_str());
      }

    } else {
      filenames.push_back (argv[argno]);
    }
  }

  if (filenames.size() == 0)
    Error ("No filenames specified.");

  if (toggle_train){

    if (gammar != GAMMA_DEFAULT)
      Error ("Gamma parameter should not be specified in training mode.");
    if (toggle_viterbi)
      Error ("The --viterbi option cannot be used in training mode.");
    if (toggle_complementary_only)
      Error ("The --complementary flag has no effect in training mode.");
    if (toggle_use_constraints)
      Error ("The --constraints flag has no effect in training mode.");
    if (toggle_partition)
      Error ("The --partition flag cannot be used in training mode.");
    if (regularization_parameter <= -1)
      Error ("Invalid regularization parameter C.  C must be greater than -1.");
    if (output_parens_filename != "")
      Error ("The --parens option cannot be used in training mode.");
    if (output_bpseq_filename != "")
      Error ("The --bpseq option cannot be used in training mode.");
    if (output_posteriors_filename != "" || posterior_cutoff != 0)
      Error ("The --posteriors option cannot be used in training mode.");
    
  } else {

    if (filenames.size() > 1 && output_parens_filename == "" &&
	output_bpseq_filename == "" && output_posteriors_filename == "")
      Error ("Output directory must be specified when performing predictions for multiple input files.");
    if (toggle_viterbi && output_posteriors_filename != "")
      Error ("The --posteriors option cannot be used with Viterbi parsing.");
    if (toggle_check)
      Error ("The --check option can only be using in training mode.");
    if (regularization_parameter != REGULARIZATION_DEFAULT)
      Error ("Regularization parameter should not be specified for prediction.");
    if (toggle_partition && toggle_viterbi)
      Error ("The --partition flag cannot be used with --viterbi.");
  }
}

