/////////////////////////////////////////////////////////////////
// InferenceEngine.h
/////////////////////////////////////////////////////////////////

#ifndef INFERENCEENGINE_H
#define INFERENCEENGINE_H

#include <queue>
#include <vector>
#include <string>
#include "Score.h"
#include "SStruct.h"
#include "Table.h"
#include "Parameters.h"

using namespace std;

/////////////////////////////////////////////////////////////////
// class InferenceEngine
/////////////////////////////////////////////////////////////////

class InferenceEngine {

  string sequence;
  int length;
  VI constraints; 
  VVI allow_range_unpaired;

  VI char_mapping;
  
  Table *inside;
  Table *outside;
  Table *posterior;
  Table *viterbi;
  Table *traceback;

  enum StateType {
    DO_OUTER,
    DO_HELIX,
    DO_HELIX1,
#ifdef PARAMS_HELIX_LENGTH
    DO_HELIX2,
    DO_HELIX3,
    DO_HELIX4,
    DO_HELIX5,
#endif
    DO_LOOP,
    DO_MULTI,
    DO_MULTI1,
    DO_MULTI2,
    NUM_STATES
  };

  // Parameters and expected counts

#ifdef PARAMS_OUTER_LENGTH
  Score score_outer_unpaired, counts_outer_unpaired;
  Score score_outer_branch, counts_outer_branch;
#endif
#ifdef PARAMS_HELIX_BASE_PAIR
  VVF score_helix_base_pair, counts_helix_base_pair;
#endif
#ifdef PARAMS_HELIX_CLOSING
  VVF score_helix_closing, counts_helix_closing;
#endif
#ifdef PARAMS_HELIX_LENGTH
  VF score_helix_length, counts_helix_length;
  Score score_helix_extend, counts_helix_extend;
#endif
#ifdef PARAMS_HELIX_STACKING
  VVVVF score_helix_stacking, counts_helix_stacking;
#endif
#ifdef PARAMS_TERMINAL_MISMATCH
  VVVVF score_terminal_mismatch, counts_terminal_mismatch;
#endif
#ifdef PARAMS_HAIRPIN_LENGTH
  VF score_hairpin_length, counts_hairpin_length;
  Score score_hairpin_extend, counts_hairpin_extend;
#endif
#ifdef PARAMS_BULGE_LENGTH
  VF score_bulge_length, counts_bulge_length;
#endif
#ifdef PARAMS_INTERNAL_FULL
  VVF score_internal_full, counts_internal_full;
#endif
#ifdef PARAMS_INTERNAL_LENGTH
  VF score_internal_length, counts_internal_length;
#endif
#ifdef PARAMS_INTERNAL_ASYMMETRY
  VF score_internal_asymmetry, counts_internal_asymmetry;
#endif
#ifdef PARAMS_MULTI_LENGTH
  Score score_multi_base, counts_multi_base;
  Score score_multi_unpaired, counts_multi_unpaired;
  Score score_multi_paired, counts_multi_paired;
#endif
#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  VVVF score_single_base_pair_stacking_left, counts_single_base_pair_stacking_left;
  VVVF score_single_base_pair_stacking_right, counts_single_base_pair_stacking_right;
#endif

  inline Score ScoreOuterUnpaired (int i) const;
  inline Score ScoreOuterBranch() const;
  inline Score ScoreHelixBasePair (int i, int j) const;
  inline Score ScoreHelixClosing (int i, int j) const;
  inline Score ScoreHelixChange (int len) const;
  inline Score ScoreHelixExtend () const;
  inline Score ScoreHelixStacking (int i1, int j1, int i2, int j2) const;
  inline Score ScoreTerminalMismatch (int i1, int j1, int i2, int j2) const;
  inline Score ScoreHairpin (int i, int j) const;
  inline Score ScoreSingle (int i, int j, int ip, int jp) const;
  inline Score ScoreMultiBase() const;
  inline Score ScoreMultiUnpaired(int i) const;
  inline Score ScoreMultiPaired () const;
  inline Score ScoreMultiMismatch (int i1, int j1, int i2, int j2) const;

  inline void CountOuterUnpaired (int i, Score value);
  inline void CountOuterBranch (Score value);
  inline void CountHelixBasePair (int i, int j, Score value);
  inline void CountHelixClosing (int i, int j, Score value);
  inline void CountHelixChange (int len, Score value);
  inline void CountHelixExtend (Score value);
  inline void CountHelixStacking (int i1, int j1, int i2, int j2, Score value);
  inline void CountTerminalMismatch (int i1, int j1, int i2, int j2, Score value);
  inline void CountHairpin (int i, int j, Score value);
  inline void CountSingle (int i, int j, int ip, int jp, Score value);
  inline void CountMultiBase (Score value);
  inline void CountMultiUnpaired (int i, Score value);
  inline void CountMultiPaired (Score value);
  inline void CountMultiMismatch (int i1, int j1, int i2, int j2, Score value);
  
  void InitializeParameters();
  void InitializeCounts();

  void TransformParameters();
  void TransformCounts();
  void WriteCounts (Parameters &parameters);

  inline void UPDATE_MAX (Score &curr_val, 
			  Score &curr_traceback,
			  const Score &new_val, 
			  const Score &new_traceback) const;
  inline Score EncodeTraceback (int i, int j) const;
  inline pair<int,int> DecodeTraceback (Score s) const;
  pair<int,pair<int,int> > make_triple (int i, int j, int k) const;

 public:

  InferenceEngine (const string &sequence);
  ~InferenceEngine ();		   

  void LoadParameters (const Parameters &parameters);
  void UseMapping (const vector<int> &mapping);
  void ComputeInside();
  void ComputeOutside();
  Score ComputePartitionCoefficient() const;
  void ComputePosterior();
  VI PredictPairingsPosterior (const double gammar) const;
  void ComputeFeatureCountExpectations(vector<double> &feature_expectations);

  void ComputeViterbi();
  VI PredictPairingsViterbi() const;

  const Table *GetInside() const;
  const Table *GetOutside() const;
  const Table *GetPosterior() const;
  const Table *GetViterbi() const;
  const Table *GetTraceback() const;
};

#endif
