/////////////////////////////////////////////////////////////////
// MakeCoords.cc
/////////////////////////////////////////////////////////////////

#include "MakeCoords.h"

using namespace std;

bool toggle_complementary_only = false;
double LEARN_RATE;

/////////////////////////////////////////////////////////////////
// MakeCoords::GetParams()
//
// Retrieve unrolled parameter vector.
/////////////////////////////////////////////////////////////////

VD MakeCoords::GetParams (const vector<Point> &points) const {
  VD res;
  for (int i = 1; i < (int) points.size(); i++){
    res.push_back (points[i].x);
    res.push_back (points[i].y);
  }
  return res;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::SetParams()
//
// Load unrolled parameter vector.
/////////////////////////////////////////////////////////////////

void MakeCoords::SetParams (const VD &values){
  coords = vector<Point>(sstruct.GetLength()+1);
  gradients = vector<Point>(sstruct.GetLength()+1);

  int k = 0;
  for (int i = 1; i < (int) coords.size(); i++){
    coords[i].x = values[k++];
    coords[i].y = values[k++];
  }
}

/////////////////////////////////////////////////////////////////
// MakeCoords::PrintParams()
//
// Print parameters.
/////////////////////////////////////////////////////////////////

void MakeCoords::PrintParams (const string &filename) const {
  ofstream outfile (filename.c_str());
  if (!outfile) Error ("Could not open coordinates file for writing.");
  for (int i = 1; i <= sstruct.GetLength(); i++){
    outfile << coords[i].x << " " << coords[i].y << endl;
  }
  outfile.close();
}

/////////////////////////////////////////////////////////////////
// ComputeLoop()
//
// Compute the residues involved in a loop.
/////////////////////////////////////////////////////////////////

VI MakeCoords::ComputeLoop (const VI &mapping, int left) const {
  VI ret (1, left);
  int level = 0;
  
  // look for other letters in the loop
  
  for (int k = left+1; k < mapping[left]; k++){
    if (mapping[k] == 0){
      if (level == 0) ret.push_back (k);
    } else {
      if (mapping[k] < k) level--;
      if (level == 0) ret.push_back (k);
      if (mapping[k] > k){
        if (level == 0) 
          for (int i = 0; i < STEM_WIDTH-1; i++) 
            ret.push_back (0);
        level++;
      }
    }
  }  
  
  ret.push_back (mapping[left]);
  for (int i = 0; i < STEM_WIDTH-1; i++) ret.push_back (0);
  
  return ret;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::ComputeLoopCenter()
//
// Given two points (x1,y1) and (x2,y2) that span a central
// angle of 2*PI*k/n (going counterclockwise), compute the
// center (xc,yc).
/////////////////////////////////////////////////////////////////

Point MakeCoords::ComputeLoopCenter (Point p1, Point p2, int k, int n) const {
  double theta = PI*(0.5 - (double)k/n);
  return Point (p1.x + 0.5 * ((p2.x-p1.x) - tan(theta) * (p2.y-p1.y)),
		p1.y + 0.5 * (tan(theta) * (p2.x-p1.x) + (p2.y-p1.y)));
}

/////////////////////////////////////////////////////////////////
// MakeCoords::ComputeLoopPositions()
//
// Given the initial and center point of a loop, and the number
// of residues in the loop, compute the positions of each residue
// in the loop (clockwise).
/////////////////////////////////////////////////////////////////

vector<Point> MakeCoords::ComputeLoopPositions (Point p1, Point center, int n) const {
  vector<Point> ret;
  double alpha = 2*PI/n;
  double dx = p1.x - center.x;
  double dy = p1.y - center.y;
  for (int i = 0; i < n; i++)
    ret.push_back(Point(center.x + cos(-alpha*i)*dx - sin(-alpha*i)*dy,
			center.y + sin(-alpha*i)*dx + cos(-alpha*i)*dy));
  return ret;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::InitialPlacement()
//
// Initial placement of all coordinates using deterministic
// procedure.
/////////////////////////////////////////////////////////////////

void MakeCoords::InitialPlacement(){
  coords = vector<Point>(sstruct.GetLength()+1);
  gradients = vector<Point>(sstruct.GetLength()+1);

  VI mapping = sstruct.GetMapping();
  const int L = (int) mapping.size() - 1;

  // set any unknown pairings to be unpaired

  for (int i = 1; i <= L; i++) mapping[i] = max (0, mapping[i]);

  // temporarily force a pairing of the first and last residues

  if (mapping[1] != 0) mapping[mapping[1]] = 0;
  if (mapping[L] != 0) mapping[mapping[L]] = 0;
  mapping[1] = L;
  mapping[L] = 1;

  // set coordinates of first and last residues

  coords[1].x = 0;
  coords[1].y = 0;
  coords[L].x = STEM_WIDTH;
  coords[L].y = 0;
  
  // iteratively find internal loops and grow a structure from them

  for (int i = 1; i <= L; i++) if (mapping[i] > i){
    VI loop = ComputeLoop (mapping, i);
    int j = mapping[i];
    Point center = ComputeLoopCenter (coords[i], coords[j], STEM_WIDTH, loop.size());
    vector<Point> positions = ComputeLoopPositions (coords[i], center, loop.size());
    for (int k = 0; k < (int) loop.size(); k++){
      if (loop[k] > 0) coords[loop[k]] = positions[k];
    }
  }
}

/////////////////////////////////////////////////////////////////
// MakeCoords::Distance()
//
// Compute squared distance between two points.
/////////////////////////////////////////////////////////////////

double MakeCoords::Distance (Point p, Point q) const {
  p.x -= q.x;
  p.y -= q.y;
  return sqrt(p.x*p.x + p.y*p.y);
}

/////////////////////////////////////////////////////////////////
// MakeCoords::AddConstraints()
//
// Add all constraints associated with loops in RNA structure.
/////////////////////////////////////////////////////////////////

void MakeCoords::AddConstraints(){
  
  VI mapping = sstruct.GetMapping();
  const int L = (int) mapping.size() - 1;

  // set any unknown pairings to be unpaired

  for (int i = 1; i <= L; i++) mapping[i] = max (0, mapping[i]);

  // add constraints for loops

  for (int i = 1; i <= L; i++) if (mapping[i] > i){
    VI loop = ComputeLoop (mapping, i);
    for (int k1 = 0; k1 < (int) loop.size(); k1++) if (loop[k1] > 0){
      for (int k2 = k1+1; k2 < (int) loop.size(); k2++) if (loop[k2] > 0){
	double dist = Distance (coords[loop[k1]], coords[loop[k2]]);
	constraints.push_back (Constraint(ConstraintType_LENGTH, loop[k1], loop[k2], 
					  dist, LOOP_STRENGTH / max(1, (int) loop.size() - 7)));
      }
    }
  }

  // add backbone constraints

  for (int i = 1; i < L; i++){
    double dist = Distance (coords[i], coords[i+1]);
    constraints.push_back (Constraint (ConstraintType_LENGTH, i, i+1, 
				       dist, BACKBONE_STRENGTH));
  }
  
  // add repulsive constraints

  for (int i = 1; i <= L; i++)
    for (int j = i+1; j <= L; j++)
      constraints.push_back (Constraint (ConstraintType_REPULSIVE, i, j, 0, REPULSIVE_STRENGTH));
}

/////////////////////////////////////////////////////////////////
// MakeCoords::CrossProduct()
//
// Compute (b-a) x (c-a).
/////////////////////////////////////////////////////////////////

double MakeCoords::CrossProduct (const Point &a, Point b, Point c) const {
  b.x -= a.x; b.y -= a.y;
  c.x -= a.x; c.y -= a.y;
  return b.x*c.y - b.y*c.x;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::IsCrossing()
//
// Return true if two line segments intersect.
/////////////////////////////////////////////////////////////////

bool MakeCoords::IsCrossing (const Point &p1, const Point &p2, const Point &q1, const Point &q2) const {
  return (CrossProduct (p1, p2, q1) * CrossProduct (p1, p2, q2) < 0 &&
	  CrossProduct (q1, q2, p1) * CrossProduct (q1, q2, p2) < 0);
}

/////////////////////////////////////////////////////////////////
// MakeCoords::IdentifyCrossingResidues()
//
// Identify all crossing residues.  Return a list of all 
// positions between crossing residues.
/////////////////////////////////////////////////////////////////

VI MakeCoords::IdentifyCrossingResidues(){

  vector<pair<int,int> > crossings;
  
  VI mapping = sstruct.GetMapping();
  const int L = (int) mapping.size() - 1;

  // set any unknown pairings to be unpaired

  for (int i = 1; i <= L; i++) mapping[i] = max (0, mapping[i]);

  for (int i = 1; i < L; i++){
    for (int j = i+2; j < L; j++){
      if (IsCrossing (coords[i], coords[i+1], coords[j], coords[j+1])){
	crossings.push_back (make_pair(i+1,j));
      }
    }
  }

  VI ret (sstruct.GetLength()+1);
  for (int i = 0; i < (int) crossings.size(); i++){
    bool good = true;
    for (int j = 0; good && j < (int) crossings.size(); j++) if (j != i)
      if (crossings[i].first <= crossings[j].first &&
	  crossings[i].second >= crossings[j].second) good = false;
    if (!good) continue;
    cerr << "Adding range: " << crossings[i].first+1 << " " << crossings[i].second-1 << endl;
    for (int j = crossings[i].first+1; j <= crossings[i].second-1; j++){
      ret[j] = 1;
    }
  }
  return ret;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::MakeCoords()
//
// Constructor.
/////////////////////////////////////////////////////////////////

MakeCoords::MakeCoords (const SStruct &sstruct) : 
  Minimizer(false), sstruct(sstruct) {
    
  InitialPlacement();
  AddConstraints();
}

/////////////////////////////////////////////////////////////////
// MakeCoords::ComputeGradient()
//
// Compute objective gradient.
/////////////////////////////////////////////////////////////////

void MakeCoords::ComputeGradient (VD &gradient, const VD &values){
  SetParams (values);
  
  for (int i = 0; i < (int) constraints.size(); i++){
    
    // precompute points involved in constraint
    
    const Point &p1 = coords[constraints[i].i];
    const Point &p2 = coords[constraints[i].j];
    const double r = Distance (p1, p2);
    
    // check type of constraint
    
    switch (constraints[i].type){
      
    case ConstraintType_LENGTH: 
      {
	double a = constraints[i].strength * constraints[i].dist;
	double b = constraints[i].strength / (2 * constraints[i].dist * constraints[i].dist);
	double dVdr = -a/(r*r) + 2*b*r;
	
	gradients[constraints[i].i].x += dVdr * (p1.x - p2.x) / r;
	gradients[constraints[i].i].y += dVdr * (p1.y - p2.y) / r;
	gradients[constraints[i].j].x += dVdr * (p2.x - p1.x) / r;
	gradients[constraints[i].j].y += dVdr * (p2.y - p1.y) / r;
      }
      break;
      
    case ConstraintType_REPULSIVE: 
      {
	double a = constraints[i].strength;
	double dVdr = -a/(r*r);
	
	gradients[constraints[i].i].x += dVdr * (p1.x - p2.x) / r;
	gradients[constraints[i].i].y += dVdr * (p1.y - p2.y) / r;
	gradients[constraints[i].j].x += dVdr * (p2.x - p1.x) / r;
	gradients[constraints[i].j].y += dVdr * (p2.y - p1.y) / r;
      }
      break;
      
    default:
      Error ("Invalid constraint type.");
    }
  }
  
  gradient = GetParams (gradients);
}

/////////////////////////////////////////////////////////////////
// MakeCoords::ComputeFunction()
//
// Compute objective function.
/////////////////////////////////////////////////////////////////

double MakeCoords::ComputeFunction (const VD &values){
  SetParams (values);

  double ret = 0;
  for (int i = 0; i < (int) constraints.size(); i++){
    
    // precompute points involved in constraint
    
    const Point &p1 = coords[constraints[i].i];
    const Point &p2 = coords[constraints[i].j];
    const double r = Distance (p1, p2);
    const double d = constraints[i].dist;

    // check type of constraint
    
    switch (constraints[i].type){
      
    case ConstraintType_LENGTH: 
      {
	double a = constraints[i].strength * constraints[i].dist;
	double b = constraints[i].strength / (2 * constraints[i].dist * constraints[i].dist);
	double V = a/r + b*r*r - (a/d + b*d*d);
	
	ret += V;
      }
      break;
      
    case ConstraintType_REPULSIVE: 
      {
	double a = constraints[i].strength;
	double V = a/r;
	
	ret += V;
      }
      break;
      
    default:
      Error ("Invalid constraint type.");
    }
  }
  
  return ret;
}

/////////////////////////////////////////////////////////////////
// MakeCoords::Report()
//
// Print feedback from Minimizer.
/////////////////////////////////////////////////////////////////

void MakeCoords::Report (const vector<double> &theta, int iteration, 
			 double objective, double step_length){
  cerr << "Minimizer: iteration " << iteration 
       << ", function value = " << objective 
       << ", step length = " << step_length << endl;
}

void MakeCoords::Report (const string &s){
  cerr << "Minimizer: " << s << endl;
}  

/////////////////////////////////////////////////////////////////
// main()
//
// Main program.
/////////////////////////////////////////////////////////////////
 
int main (int argc, char **argv){

  if (argc != 3){
    cerr << "Usage: " << argv[0] << " BPSEQFILE OUTFILE" << endl
	 << endl
	 << "       where BPSEQFILE    is the name of the input BPSEQ file" << endl
	 << "             OUTFILE      is the name of the output coordinates file" << endl;
    exit (1);
  }

  SStruct sstruct (argv[1]);
  MakeCoords folder (sstruct);
  VD theta = folder.GetParams(folder.coords);
  folder.LBFGS (theta);
  folder.SetParams (theta);
  folder.PrintParams (argv[2]);
}
