/////////////////////////////////////////////////////////////////
// Parameters.cc
/////////////////////////////////////////////////////////////////

#include "Parameters.h"
#include "Defaults.h"

/////////////////////////////////////////////////////////////////
// ParameterGroup::ParameterGroup()
// 
// Default constructor.
/////////////////////////////////////////////////////////////////

ParameterGroup::ParameterGroup() {}

/////////////////////////////////////////////////////////////////
// ParameterGroup::ParameterGroup()
// 
// Constructor.
/////////////////////////////////////////////////////////////////

ParameterGroup::ParameterGroup (const string &name, int begin, int end) :
  name(name), begin(begin), end(end) {}

/////////////////////////////////////////////////////////////////
// ParameterGroup::ParameterGroup()
// 
// Copy constructor.
/////////////////////////////////////////////////////////////////

ParameterGroup::ParameterGroup (const ParameterGroup &rhs) :
  name(rhs.name), begin(rhs.begin), end(rhs.end) {}

/////////////////////////////////////////////////////////////////
// Parameters::Parameters()
//
// Constructor.
/////////////////////////////////////////////////////////////////

Parameters::Parameters() : group_begin(-1) {

#ifdef PARAMS_OUTER_LENGTH
  BeginGroup("outer_length");
  NewParameter("outer_unpaired");
  RegularizeToZero("outer_unpaired");
  NewParameter("outer_branch");
  RegularizeToZero("outer_branch");
  EndGroup();
#endif

#ifdef PARAMS_HELIX_BASE_PAIR
  BeginGroup("helix_base_pair");
  for (int i = 0; i < M; i++){
    for (int j = i; j < M; j++){
      string name = SPrintF ("helix_base_pair_%c%c", alphabet[i], alphabet[j]);
      NewParameter(name);
      RegularizeToZero(name);
      if (i < j){
	string name2 = SPrintF ("helix_base_pair_%c%c", alphabet[j], alphabet[i]);
	AddParameterAlias (name2, name);
      }
    }
  }
  EndGroup();
#endif

#ifdef PARAMS_HELIX_CLOSING
  BeginGroup("helix_closing");
  for (int i = 0; i < M; i++){
    for (int j = 0; j < M; j++){
      string name = SPrintF ("helix_closing_%c%c", alphabet[i], alphabet[j]);
      NewParameter(name);
      RegularizeToZero(name);
    }
  }
  EndGroup();
#endif

#ifdef PARAMS_HELIX_STACKING
  BeginGroup("helix_stacking");
  
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  string name = SPrintF ("helix_stacking_%c%c%c%c",
				 alphabet[i1], alphabet[j1],
				 alphabet[i2], alphabet[j2]);
	  string name2 = SPrintF ("helix_stacking_%c%c%c%c",
				  alphabet[j2], alphabet[i2],
				  alphabet[j1], alphabet[i1]);

	  if (name < name2){
	    NewParameter(name);
	    RegularizeToZero(name);
	    AddParameterAlias(name2,name);
	  } else if (name == name2){
	    NewParameter(name);
	    RegularizeToZero(name);
	  }
	}
      }
    }
  }

  EndGroup();
#endif

#ifdef PARAMS_HELIX_LENGTH
  BeginGroup("helix_length");
  for (int len = 1; len <= D; len++){
    string name = SPrintF ("helix_length_%d", len);
    NewParameter(name);
    RegularizeToZero(name);
  }
  EndGroup();

  BeginGroup("helix_extend");
  NewParameter("helix_extend");
  RegularizeToZero("helix_extend");
  EndGroup();
#endif

#ifdef PARAMS_TERMINAL_MISMATCH
  BeginGroup("terminal_mismatch");
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  string name = SPrintF ("terminal_mismatch_%c%c%c%c",
				 alphabet[i1], alphabet[j1],
				 alphabet[i2], alphabet[j2]);
	  NewParameter(name);
	  RegularizeToZero(name);
	}
      }
    }
  }
  EndGroup();
#endif

#ifdef PARAMS_HAIRPIN_LENGTH
  BeginGroup("hairpin_length");
  for (int i = 0; i <= C; i++){
    string name = SPrintF ("hairpin_length_%d", i);
    NewParameter (name);
    RegularizeToZero (name);
  }
  EndGroup();

  BeginGroup("hairpin_affine");
  NewParameter ("hairpin_extend");
  RegularizeToZero("hairpin_extend");
  EndGroup();
#endif

#ifdef PARAMS_BULGE_LENGTH
  BeginGroup("bulge_length");
  for (int i = 0; i <= C; i++){
    string name = SPrintF ("bulge_length_%d", i);
    NewParameter (name);
    RegularizeToZero (name);
  }
  EndGroup();
#endif

#ifdef PARAMS_INTERNAL_FULL
  BeginGroup("internal_full");
  for (int i = 1; i <= C; i++){
    for (int j = i; i+j <= C; j++){
      string name = SPrintF ("internal_full_%d_%d", i, j);
      NewParameter (name);
      RegularizeToZero (name);
      if (i < j){
	string name2 = SPrintF ("internal_full_%d_%d", j, i);
	AddParameterAlias (name2, name);
      }
    }
  }
  EndGroup();
#endif

#ifdef PARAMS_INTERNAL_LENGTH
  BeginGroup("internal_length");
  for (int i = 1; i <= C; i++){
    string name = SPrintF ("internal_length_%d", i);
    NewParameter (name);
    RegularizeToZero (name);
  }
  EndGroup();
#endif

#ifdef PARAMS_INTERNAL_ASYMMETRY
  BeginGroup("internal_asymmetry");
  for (int i = 0; i <= C; i++){
    string name = SPrintF ("internal_asymmetry_%d", i);
    NewParameter (name);
    RegularizeToZero (name);
  }
  EndGroup();
#endif

#ifdef PARAMS_MULTI_LENGTH
  BeginGroup("multi_length");
  NewParameter ("multi_base");
  RegularizeToZero("multi_base");
  NewParameter ("multi_unpaired");
  RegularizeToZero("multi_unpaired");
  NewParameter ("multi_paired");
  RegularizeToZero("multi_paired");
  EndGroup();
#endif

#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  BeginGroup("single_base_pair_stacking_left");
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	string name = SPrintF ("single_base_pair_stacking_left_%c%c%c",
			       alphabet[i1], alphabet[j1], alphabet[i2]);
	NewParameter(name);
	RegularizeToZero(name);
      }
    }
  }
  EndGroup();

  BeginGroup("single_base_pair_stacking_right");
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	string name = SPrintF ("single_base_pair_stacking_right_%c%c%c",
			       alphabet[i1], alphabet[j1], alphabet[i2]);
	NewParameter(name);
	RegularizeToZero(name);
      }
    }
  }
  EndGroup();
#endif

}

/////////////////////////////////////////////////////////////////
// Parameters::BeginGroup()
//
// Mark the beginning of a new parameter group.
/////////////////////////////////////////////////////////////////

void Parameters::BeginGroup (const string &name){
  Assert (group_begin == -1, "Already started parameter group.");
  group_begin = values.size();
  group_name = name;
}

/////////////////////////////////////////////////////////////////
// Parameters::EndGroup()
//
// Mark the end of a new parameter group.
/////////////////////////////////////////////////////////////////

void Parameters::EndGroup (){
  Assert (group_begin != -1, "No currently open parameter group.");
  param_groups.push_back (ParameterGroup(group_name,group_begin,values.size()-1));
  group_begin = -1;
}

/////////////////////////////////////////////////////////////////
// Parameters::NewParameter()
//
// Create new parameter.
/////////////////////////////////////////////////////////////////

void Parameters::NewParameter(const string &s){
  map<string,int>::iterator iter = parameter_name_to_index.find (s);
  Assert (iter == parameter_name_to_index.end(), "Attempt to create parameter with duplicate name.");

  parameter_name_to_index[s] = values.size();
  values.push_back (0);
  names.push_back (s);
}

/////////////////////////////////////////////////////////////////
// Parameters::AddParameterAlias()
//
// Create new alias for parameter.
/////////////////////////////////////////////////////////////////

void Parameters::AddParameterAlias (const string &alias, const string &name){
  map<string,int>::iterator iter = parameter_name_to_index.find (alias);
  Assert (iter == parameter_name_to_index.end(), "Attempt to create parameter with duplicate name.");
  map<string,int>::iterator iter2 = parameter_name_to_index.find (name);
  Assert (iter2 != parameter_name_to_index.end(), "Request for unknown parameter.");
  parameter_name_to_index[alias] = iter2->second;
}

/////////////////////////////////////////////////////////////////
// Parameters::InitializeRandomly()
//
// Load small random values into Parameter object.
/////////////////////////////////////////////////////////////////

void Parameters::InitializeRandomly (){
  srand48(time(0));
  for (int i = 0; i < (int) values.size(); i++){
#ifdef SANITY
    values[i] = 0;
#else
    values[i] = drand48() * 1e-10 - 5e-11;
#endif
  }
}

/////////////////////////////////////////////////////////////////
// Parameters::Clear()
//
// Zero out parameter vector.
/////////////////////////////////////////////////////////////////

void Parameters::Clear(){
  for (int i = 0; i < (int) values.size(); i++)
    values[i] = 0;
}

/////////////////////////////////////////////////////////////////
// Parameters::GetNumValues()
//
// Retrieve number of parameters from Parameter object.
/////////////////////////////////////////////////////////////////

int Parameters::GetNumValues() const {
  return (int) values.size();
}

/////////////////////////////////////////////////////////////////
// Parameters::GetValues()
//
// Retrieve values from Parameter object.
/////////////////////////////////////////////////////////////////

const vector<double> Parameters::GetValues() const {
  vector<double> ret;
  for (int i = 0; i < (int) values.size(); i++)
    ret.push_back ((double) values[i]);
  return ret;
}

/////////////////////////////////////////////////////////////////
// Parameters::LoadValues()
//
// Load values into Parameter object.
/////////////////////////////////////////////////////////////////

Parameters &Parameters::LoadValues (const vector<double> &values){
  Assert (this->values.size() == values.size(), "Parameter vector of incorrect dimension.");
  this->values = values;
  return *this;
}

/////////////////////////////////////////////////////////////////
// Parameters::ReadFromFile()
//
// Read parameters from file.
/////////////////////////////////////////////////////////////////

void Parameters::ReadFromFile (const string &filename){
  vector<string> read_names;
  double value;
  string name;
  string s;
  
  ifstream infile (filename.c_str());
  if (infile.fail()) Error (("Could not open file \"" + filename + "\" for reading.").c_str());
  values.clear();
  while (getline (infile, s)){
    if (s.length() == 0 || s[0] == '#') continue;
    istringstream iss (s);
    if (iss >> name >> value){
      read_names.push_back (name);
      values.push_back (value);
    }
  }
  infile.close();
  
  // check to make sure that read parameter vector is valid
  
  if (read_names.size() < names.size()) Error ("Parameter file contains too few parameters.");
  if (read_names.size() > names.size()) Error ("Parameter file contains too many parameters.");
  if (read_names != names) Error ("Unexpected parameter names in parameter file.");
}

/////////////////////////////////////////////////////////////////
// Parameters::WriteToFile()
//
// Write parameters to file.
/////////////////////////////////////////////////////////////////

void Parameters::WriteToFile (const string &filename){
  ofstream outfile (filename.c_str());
  if (outfile.fail()) Error (("Could not open file \"" + filename + "\" for writing.").c_str());
  for (int i = 0; i < (int) values.size(); i++)
    outfile << names[i] << " " << setprecision(10) << values[i] << endl;
  outfile.close();
}

/////////////////////////////////////////////////////////////////
// Parameters::GetParam()
//
// Retrieve the value for a particular parameter.
/////////////////////////////////////////////////////////////////

Score Parameters::GetParam (const string &s) const {
  map<string,int>::const_iterator iter = parameter_name_to_index.find (s);
  Assert (iter != parameter_name_to_index.end(), "Request for unknown parameter.");
  return values[iter->second];
}

/////////////////////////////////////////////////////////////////
// Parameters::AccumulateCounts()
//
// Find location for a particular parameter and add counts to
// that location.
/////////////////////////////////////////////////////////////////

void Parameters::AccumulateCounts (const string &s, Score value){
  map<string,int>::const_iterator iter = parameter_name_to_index.find (s);
  Assert (iter != parameter_name_to_index.end(), "Request for unknown parameter.");
  values[iter->second] += value;
}

/////////////////////////////////////////////////////////////////
// Parameters::ComputeRegularizationPenalty()
//
// Compute regularization penalty.
/////////////////////////////////////////////////////////////////

double Parameters::ComputeRegularizationPenalty () const {
  double ret = 0;
  for (int i = 0; i < (int) regularization_list.size(); i++){
    int j = regularization_list[i].first;
    int k = regularization_list[i].second;
    double first_val = (j == -1) ? 0 : values[j];
    double second_val = (k == -1) ? 0 : values[k];
#ifdef HUBER_HINGE
    double diff = fabs (first_val - second_val);
    if (diff > HUBER_HINGE){
      ret += diff * diff;
    } else {
      ret += HUBER_HINGE * (2 * diff - HUBER_HINGE);
    }
#else
    ret += (first_val - second_val) * (first_val - second_val);
#endif
  }
  return ret;
}

/////////////////////////////////////////////////////////////////
// Parameters::ComputeRegularizationGradient()
//
// Compute regularization gradient.
/////////////////////////////////////////////////////////////////

const vector<double> Parameters::ComputeRegularizationGradient () const {
  vector<double> ret (values.size());
  for (int i = 0; i < (int) regularization_list.size(); i++){
    int j = regularization_list[i].first;
    int k = regularization_list[i].second;
    double first_val = (j == -1) ? 0 : values[j];
    double second_val = (k == -1) ? 0 : values[k];
#ifdef HUBER_HINGE
    double diff = fabs (first_val - second_val);
    if (diff > HUBER_HINGE){
      if (j != -1) ret[j] += 2 * (first_val - second_val > 0 ? HUBER_HINGE : -HUBER_HINGE);
      if (k != -1) ret[k] += 2 * (second_val - first_val > 0 ? HUBER_HINGE : -HUBER_HINGE);
    } else {
      if (j != -1) ret[j] += 2 * (first_val - second_val);
      if (k != -1) ret[k] += 2 * (second_val - first_val);
    }    
#else
    if (j != -1) ret[j] += 2 * (first_val - second_val);
    if (k != -1) ret[k] += 2 * (second_val - first_val);
#endif
  }
  return ret;
}

/////////////////////////////////////////////////////////////////
// Parameters::RegularizeToZero()
//
// Regularize a parameter to zero.
/////////////////////////////////////////////////////////////////

void Parameters::RegularizeToZero (const string &s){
  map<string,int>::const_iterator iter = parameter_name_to_index.find (s);
  Assert (iter != parameter_name_to_index.end(), "Request for unknown parameter.");

  if (find (regularization_list.begin(),
            regularization_list.end(),
            make_pair(-1, iter->second)) !=
      regularization_list.end()) return;

  regularization_list.push_back (make_pair(-1, iter->second));
}

/////////////////////////////////////////////////////////////////
// Parameters::RegularizeTogether()
//
// Regularize two parameters to be close to each other
/////////////////////////////////////////////////////////////////

void Parameters::RegularizeTogether (const string &s, const string &t){
  map<string,int>::const_iterator iter = parameter_name_to_index.find (s);
  Assert (iter != parameter_name_to_index.end(), "Request for unknown parameter.");
  map<string,int>::const_iterator iter2 = parameter_name_to_index.find (t);
  Assert (iter2 != parameter_name_to_index.end(), "Request for unknown parameter.");
  
  if (find (regularization_list.begin(),
	    regularization_list.end(),
	    make_pair(iter->second, iter2->second)) !=
      regularization_list.end()) return;

  if (find (regularization_list.begin(),
	    regularization_list.end(),
	    make_pair(iter2->second, iter->second)) !=
      regularization_list.end()) return;
  
  regularization_list.push_back (make_pair(iter->second, iter2->second));
}

/////////////////////////////////////////////////////////////////
// Parameters::GetParameterGroups()
//
// Retrieve parameter groups.
/////////////////////////////////////////////////////////////////

const vector<ParameterGroup> &Parameters::GetParameterGroups() const {
  return param_groups;
}
