/////////////////////////////////////////////////////////////////
// Score.h
/////////////////////////////////////////////////////////////////

#ifndef SCORE_H
#define SCORE_H

#include <cmath>
#include <iostream>
#include <iomanip>
#include <vector>
#include <limits>

using namespace std;

/////////////////////////////////////////////////////////////////
// Types
/////////////////////////////////////////////////////////////////

#ifdef SANITY
#define EXACT
#endif

#ifdef EXACT
typedef double Score;
#else
typedef float Score;
#endif

/////////////////////////////////////////////////////////////////
// Constants
/////////////////////////////////////////////////////////////////

const Score NaN = numeric_limits<Score>::quiet_NaN();
const Score LOG_ZERO = (Score) -2e20;
const Score LOG_ZERO_REDUCED = LOG_ZERO/2;
const Score LOG_ONE = (Score) 0;

/////////////////////////////////////////////////////////////////
// LOG()
//
// Compute log(x).
/////////////////////////////////////////////////////////////////

inline Score LOG (Score x){
#ifdef EXACT
  return log(x);
#else
  /*
  union {
    float f;
    int i;
  } v;

  v.f = x;
  const int log_2 = ((v.i >> 23) & 255) - 128;
  v.i &= ~(255 << 23);
  v.i += 127 << 23;
  v.f = ((-1.0f/3) * v.f + 2) * v.f - 2.0f/3;
  
  return (v.f + log_2) * 0.69314718f;
  */
  return logf(x);
#endif
}

/////////////////////////////////////////////////////////////////
// EXP()
//
// Compute exp(x).
/////////////////////////////////////////////////////////////////

inline Score EXP (Score x){

#ifdef EXACT
  return exp(x);
#else

  /*
  if (x < -16) return 0;
  
  union {
    float d;
    int i;
  } eco;
  eco.i = (int)((1<<23)/M_LN2*x + ((1<<23)*127 - 486412));

  return eco.d;
  */
  
  if (x > (Score) -2){
    if (x > (Score) -0.5){
      if (x > (Score) 0)
	return expf(x);
      return (((((Score) 0.03254409303190190000*x + 
		 (Score) 0.16280432765779600000)*x + 
		(Score) 0.49929760485974900000)*x + 
	       (Score) 0.99995149601363700000)*x + 
	      (Score) 0.99999925508501600000);
    }
    if (x > (Score) -1)
      return (((((Score) 0.01973899026052090000*x +
		 (Score) 0.13822379685007000000)*x + 
		(Score) 0.48056651562365000000)*x + 
	       (Score) 0.99326940370383500000)*x + 
	      (Score) 0.99906756856399500000);
    return (((((Score) 0.00940528203591384000*x + 
	       (Score) 0.09414963667859410000)*x + 
	      (Score) 0.40825793595877300000)*x + 
	     (Score) 0.93933625499130400000)*x + 
	    (Score) 0.98369508190545300000);
  }
  if (x > (Score) -8){
    if (x > (Score) -4)
      return (((((Score) 0.00217245711583303000*x + 
		 (Score) 0.03484829428350620000)*x + 
		(Score) 0.22118199801337800000)*x + 
	       (Score) 0.67049462206469500000)*x + 
	      (Score) 0.83556950223398500000);
    return (((((Score) 0.00012398771025456900*x + 
	       (Score) 0.00349155785951272000)*x + 
	      (Score) 0.03727721426017900000)*x + 
	     (Score) 0.17974997741536900000)*x + 
	    (Score) 0.33249299994217400000);
  }
  if (x > (Score) -16)
    return (((((Score) 0.00000051741713416603*x + 
	       (Score) 0.00002721456879608080)*x + 
	      (Score) 0.00053418601865636800)*x + 
	     (Score) 0.00464101989351936000)*x + 
	    (Score) 0.01507447981459420000);
  return 0;

#endif
}

/////////////////////////////////////////////////////////////////
// LOOKUP()
//
// Compute log(1 + exp(x)).
/////////////////////////////////////////////////////////////////

inline Score LOOKUP (Score x){
#ifdef EXACT
  if (x > (Score) 50) return x;
  if (x < (Score) -50) return (Score) 0;
  return log(exp(x)+1);
#else
  
  if (x < (Score) 2){
    if (x < (Score) 0.5){
      if (x < (Score) 0)
	return logf (expf(x) + 1);
      
      return (((((Score) -0.00486373205785640000*x - 
		 (Score) 0.00020245408813934800)*x +
		(Score) 0.12504222666029800000)*x + 
	       (Score) 0.49999685320563000000)*x + 
	      (Score) 0.69314723138948900000);
    }
    if (x < 1)
      return (((((Score) -0.00278634205460548000*x - 
		 (Score) 0.00458097251248546000)*x + 
		(Score) 0.12865849880472500000)*x + 
	       (Score) 0.49862228499205200000)*x + 
	      (Score) 0.69334810088688000000);
    return (((((Score) 0.00059633755154209200*x - 
	       (Score) 0.01918996666063320000)*x + 
	      (Score) 0.15288232492093800000)*x + 
	     (Score) 0.48039958825756900000)*x + 
	    (Score) 0.69857578503189200000);
  }
  if (x < 8){
    if (x < 4)
      return (((((Score) 0.00135958539181047000*x -
		 (Score) 0.02329807659316430000)*x + 
		(Score) 0.15885799609532100000)*x + 
	       (Score) 0.48167498563270800000)*x + 
	      (Score) 0.69276185058669200000);
    return (((((Score) 0.00011992394456683500*x - 
	       (Score) 0.00338464503306568000)*x + 
	      (Score) 0.03622746366545470000)*x + 
	     (Score) 0.82481250248383700000)*x + 
	    (Score) 0.32507892994863100000);
  }
  if (x < 16)
    return (((((Score) 0.00000051726300753785*x -
	       (Score) 0.00002720671238876090)*x + 
	      (Score) 0.00053403733818413500)*x + 
	     (Score) 0.99536021775747900000)*x + 
	    (Score) 0.01507065715532010000);
  return x;
#endif
}

/////////////////////////////////////////////////////////////////
// LOG_ADD()
//
// Compute log(exp(x) + exp(y)).
/////////////////////////////////////////////////////////////////

inline Score LOG_ADD (Score x, Score y){
  if (x < y){ Score t = x; x = y; y = t; }
  if (y <= LOG_ZERO_REDUCED) return x;
  return LOOKUP(x-y) + y;
}

/////////////////////////////////////////////////////////////////
// LOG_PLUS_EQUALS()
//
// Set x = log(exp(x) + exp(y)).
/////////////////////////////////////////////////////////////////

inline void LOG_PLUS_EQUALS (Score &x, Score y){
  if (x < y){ Score t = x; x = y; y = t; }
  if (y > LOG_ZERO_REDUCED) 
    x = LOOKUP(x-y) + y;
}

#endif
