/////////////////////////////////////////////////////////////////
// Contrafold.cpp
/////////////////////////////////////////////////////////////////

// Include files

#ifdef MULTI
#include <mpi.h>
#endif

#include <cerrno>
#include "Utilities.hpp"
#include "OptimizationWrapper.hpp"

// Constants

const double GAMMA_DEFAULT = 6;
const double REGULARIZATION_DEFAULT = 0;

// Global variables

double gammar = GAMMA_DEFAULT;
double posterior_cutoff = 0;
bool toggle_check = false;
bool toggle_train = false;
bool toggle_sanity = false;
bool toggle_verbose = false;
bool toggle_viterbi = false;
bool toggle_partition = false;
bool toggle_complementary_only = true;
bool toggle_use_constraints = false;
double holdout_ratio = -1;
std::vector<double> regularization_parameters;

std::vector<std::string> filenames;
std::string params_filename = "";
std::string output_parens_destination = "";
std::string output_bpseq_destination = "";
std::string output_posteriors_destination = "";

// Function prototypes

void Usage();
void Version();
void ParseParameters(int argc, char **argv);
void RunSanityCheck(const std::vector<std::string> &filenames);
void RunTrainingMode(const std::vector<std::string> &filenames,
                     const std::string &output_parens_destination,
                     const std::string &output_bpseq_destination,
                     const std::string &output_posteriors_destination);
void RunPredictionMode(const std::vector<std::string> &filenames,
                       const std::string &output_parens_destination,
                       const std::string &output_bpseq_destination,
                       const std::string &output_posteriors_destination);

/////////////////////////////////////////////////////////////////
// main()
//
// Main program.
/////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
#ifdef MULTI
    MPI_Init(&argc, &argv);
#endif
    
    // First, parse parameters
    
    ParseParameters(argc, argv);
    
    // Perform either training or prediction.
    
    if (toggle_train)
    {
        if (toggle_sanity)
            RunSanityCheck(filenames);
        else
            RunTrainingMode(filenames,
                            output_parens_destination,
                            output_bpseq_destination,
                            output_posteriors_destination);
    }
    else
    {
        RunPredictionMode(filenames,
                          output_parens_destination,
                          output_bpseq_destination,
                          output_posteriors_destination);
    }
    
#ifdef MULTI
    MPI_Finalize();
#endif
}

/////////////////////////////////////////////////////////////////
// Usage()
//
// Display program usage.
/////////////////////////////////////////////////////////////////

void Usage()
{
    Parameters params;
    int num_groups = int(params.GetNumHyperparameterGroups());
    
    std::cerr << std::endl
              << "Usage: contrafold [predict|train] [OPTION]... INFILE(s)" << std::endl 
              << std::endl
              << "       where [OPTION]...   is a list of zero or more optional arguments" << std::endl
              << "             INFILE(s)     is the name of the input BPSEQ, parenthesized, or FASTA file(s)" << std::endl 
              << std::endl
              << "Miscellaneous arguments:" << std::endl
              << "  --version                display program version information" << std::endl
              << "  --verbose                show detailed console output" << std::endl
              << "  --params FILENAME        use particular model parameters" << std::endl
              << std::endl 
              << "Additional arguments for 'predict' mode:" << std::endl
              << "  --gamma GAMMA            set sensivity/specificity tradeoff parameter (default: GAMMA=" << GAMMA_DEFAULT << ")" << std::endl
              << "                             if GAMMA > 1, emphasize sensitivity" << std::endl
              << "                             if 0 <= GAMMA <= 1, emphasize specificity" << std::endl
              << "                             if GAMMA < 0, try tradeoff parameters of 2^k for k = -5,...,10" << std::endl
              << "  --viterbi                use Viterbi instead of posterior decoding" << std::endl
              << "  --noncomplementary       allow non-{AU,CG,GU} pairs" << std::endl
              << "  --constraints            use existing constraints (requires BPSEQ or parenthesized format input)" << std::endl
              << std::endl
              << "  --parens OUTFILEORDIR    write parenthesized output to file or directory" << std::endl
              << "  --bpseq OUTFILEORDIR     write BPSEQ output to file or directory" << std::endl
              << "  --posteriors CUTOFF OUTFILEORDIR" << std::endl
              << "                           write posterior pairing probabilities to file or directory" << std::endl
              << "  --partition              compute the partition function only" << std::endl
              << std::endl
              << "Additional arguments for training (many input files may be specified):" << std::endl
              << "  --sanity                 perform gradient sanity check" << std::endl
              << "  --check                  check only for valid parses" << std::endl
              << "  --holdout F              use fraction F of training data for holdout cross-validation" << std::endl
              << "  --regularize C1 ... C" << num_groups << (num_groups < 10 ? " " : "")
              << "  perform BFGS training, using regularization with coefficients C1 ... C" << num_groups << std::endl
              << std::endl;
    exit(0);
}

/////////////////////////////////////////////////////////////////
// Version()
//
// Display program version.
/////////////////////////////////////////////////////////////////

void Version()
{
    std::cerr << "CONTRAFold version 2.00 - RNA secondary structure prediction" << std::endl << std::endl
              << "Written by Chuong B. Do" << std::endl;
    exit(0);
}

/////////////////////////////////////////////////////////////////
// ParseParameters()
//
// Parse command line parameters.
/////////////////////////////////////////////////////////////////

void ParseParameters(int argc, char **argv)
{
    if (argc < 2) Usage();
    filenames.clear();
    
    // Check for prediction or training mode
    
    if (!strcmp(argv[1], "train"))
        toggle_train = true;
    else
        if (strcmp(argv[1], "predict"))
            Error("CONTRAfold must be run in either 'predict' or 'train' mode.");
    
    for (int argno = 2; argno < argc; argno++)
    {
        if (argv[argno][0] == '-')
        {
            
            // Miscellaneous options
            
            if (!strcmp(argv[argno], "--version"))
            {
                Version();
            }
            else if (!strcmp(argv[argno], "--verbose"))
            {
                toggle_verbose = true;
            }
            else if (!strcmp(argv[argno], "--params"))
            {
                if (argno == argc - 1) Error("Must specify FILENAME after --params.");
                params_filename = argv[++argno];
            }
            else if (!strcmp(argv[argno], "--partition"))
            {
                toggle_partition = true;
            }

            // Prediction options
            
            else if (!strcmp(argv[argno], "--gamma"))
            {
                if (argno == argc - 1) Error("Must specify trade-off parameter GAMMA after --gamma.");
                if (!ConvertToNumber(argv[++argno], gammar))
                    Error("Unable to parse value after --gamma.");
            }
            else if (!strcmp(argv[argno], "--viterbi"))
            {
                toggle_viterbi = true;
            }
            else if (!strcmp(argv[argno], "--noncomplementary"))
            {
                Error("Sorry, the --noncomplementary flag is currently not available!");
                toggle_complementary_only = false;
            }
            else if (!strcmp(argv[argno], "--constraints"))
            {
                toggle_use_constraints = true;
            }
            else if (!strcmp(argv[argno], "--parens"))
            {
                if (argno == argc - 1) Error("Must specify output file or directory name after --parens.");
                output_parens_destination = argv[++argno];
            }
            else if (!strcmp(argv[argno], "--bpseq"))
            {
                if (argno == argc - 1) Error("Must specify output file or directory name after --bpseq.");
                output_bpseq_destination = argv[++argno];
            }
            else if (!strcmp(argv[argno], "--posteriors"))
            {
                if (argno == argc - 1) Error("Must specify posterior probability threshold CUTOFF after --posteriors.");
                if (!ConvertToNumber(argv[++argno], posterior_cutoff))
                    Error("Unable to parse cutoff value after --posteriors.");
                if (argno == argc - 1) Error("Must specify output file or directory for --posteriors.");
                output_posteriors_destination = argv[++argno];
            }
            
            // Training options
            
            else if (!strcmp(argv[argno], "--sanity"))
            {
                toggle_sanity = true;
            }
            else if (!strcmp(argv[argno], "--check"))
            {
                toggle_check = true;
            }
            else if (!strcmp(argv[argno], "--holdout"))
            {
                if (argno == argc - 1) Error("Must specify holdout ratio F after --regularize.");
                if (!ConvertToNumber(argv[++argno], holdout_ratio))
                    Error("Unable to parse holdout ratio.");
                if (holdout_ratio < 0 || holdout_ratio > 1)
                    Error("Holdout ratio must be between 0 and 1.");
            }
            else if (!strcmp(argv[argno], "--regularize"))
            {
                Parameters params;
                if (argno + int(params.GetNumHyperparameterGroups()) >= argc) Error("Must specify regularization parameters C1 [C2...] after --regularize.");
                regularization_parameters.resize(params.GetNumHyperparameterGroups());
                for (size_t i = 0; i < params.GetNumHyperparameterGroups(); i++)
                {
                    if (!ConvertToNumber(argv[++argno], regularization_parameters[i]))
                        Error("Unable to parse regularization parameters after --regularize.");
                    if (regularization_parameters[i] < 0)
                        Error("Regularization parameter should not be negative.");
                }
            }
            else
            {
                Error(("Unknown option \"" + std::string(argv[argno]) + "\" specified.  Run program without any arguments to see command-line options.").c_str());
            }
            
        }
        else
        {
            filenames.push_back(argv[argno]);
        }
    }
    
    if (filenames.size() == 0)
        Error("No filenames specified.");
    
    if (toggle_train)
    {
        if (gammar != GAMMA_DEFAULT)
            Error("Gamma parameter should not be specified in training mode.");
        if (toggle_use_constraints)
            Error("The --constraints flag has no effect in training mode.");
        if (toggle_partition)
            Error("The --partition flag cannot be used in training mode.");
        if (output_parens_destination != "")
            Error("The --parens option cannot be used in training mode.");
        if (output_bpseq_destination != "")
            Error("The --bpseq option cannot be used in training mode.");
        if (output_posteriors_destination != "" || posterior_cutoff != 0)
            Error("The --posteriors option cannot be used in training mode.");
        if (regularization_parameters.size() > 0 && holdout_ratio > 0)
            Error("The --holdout and --regularize options cannot be specified simultaneously.");
    }
    else
    {
        if (gamma < 0 && output_parens_destination == "" &&
            output_bpseq_destination == "" && output_posteriors_destination == "")
            Error("Output directory must be specified when using GAMMA < 0.");
#ifdef MULTI
        if (filenames.size() > 1 && output_parens_destination == "" &&
            output_bpseq_destination == "" && output_posteriors_destination == "")
            Error("Output directory must be specified when performing predictions for multiple input files.");
#endif
        if (toggle_viterbi && output_posteriors_destination != "")
            Error("The --posteriors option cannot be used with Viterbi parsing.");
        if (toggle_check)
            Error("The --check option can only be using in training mode.");
        if (toggle_partition && toggle_viterbi)
            Error("The --partition flag cannot be used with --viterbi.");
        if (toggle_sanity)
            Error("The --sanity flag can only be used in training mode.");
    }
}

/////////////////////////////////////////////////////////////////
// RunSanityCheck()
//
// Perform CONTRAfold gradient sanity check.
/////////////////////////////////////////////////////////////////

void RunSanityCheck(const std::vector<std::string> &filenames)
{
    // create a "computation" object which will handle all requests for
    // performing heavy-duty computations (e.g., parsing sequences, computing
    // logloss gradients, etc).
    
    Computation computation(filenames, "", "", "", 
                            toggle_complementary_only,
                            toggle_use_constraints, 
                            toggle_partition, 
                            toggle_viterbi,
                            toggle_verbose,
                            posterior_cutoff);
    
    if (computation.IsComputeNode())
    {
        computation.RunComputeNode();
        return;
    }
    
    // retrieve base parameter values
    
    Parameters params;
    std::vector<double> values =
        (params_filename != "") ?
        params.ReadFromFile(params_filename) :
        std::vector<double>(params.GetNumParameters(), 0.0);
    
    OptimizationWrapper optimizer(computation, params, values);
    
    computation.SanityCheckGradient(computation.GetAllUnits(), values);
    computation.StopComputeNodes();
}

/////////////////////////////////////////////////////////////////
// RunTrainingMode()
//
// Run CONTRAfold in training mode.
/////////////////////////////////////////////////////////////////

void RunTrainingMode(const std::vector<std::string> &filenames,
                     const std::string &output_parens_destination,
                     const std::string &output_bpseq_destination,
                     const std::string &output_posteriors_destination)
{
    // create a "computation" object which will handle all requests for
    // performing heavy-duty computations (e.g., parsing sequences, computing
    // logloss gradients, etc).
    
    Computation computation(filenames,
                            output_parens_destination,
                            output_bpseq_destination,
                            output_posteriors_destination,
                            toggle_complementary_only,
                            toggle_use_constraints, 
                            toggle_partition, 
                            toggle_viterbi,
                            toggle_verbose,
                            posterior_cutoff);
    
    if (computation.IsComputeNode())
    {
        computation.RunComputeNode();
        return;
    }
    
    // retrieve base parameter values
    
    Parameters params;
    std::vector<double> values =
        (params_filename != "") ?
        params.ReadFromFile(params_filename) :
        std::vector<double>(params.GetNumParameters());
    
    OptimizationWrapper optimizer(computation, params, values);
    
    // get initial parameter values
    
    values = params.GetRandomValues();
    
    // filter work units for nonparsable examples
    
    std::vector<int> units = computation.FilterNonparsable(computation.GetAllUnits(), values);
    
    // decide between using a fixed regularization parameter or
    // using cross-validation to determine regularization parameters

    if (holdout_ratio < 0)
    {
        if (regularization_parameters.size() == 0)
            regularization_parameters.resize(params.GetNumHyperparameterGroups(), REGULARIZATION_DEFAULT);
        optimizer.Train(units, values, regularization_parameters, toggle_viterbi);
    }
    else
    {
        optimizer.LearnHyperparameters(units, values, holdout_ratio, toggle_viterbi);
    }
    
    params.WriteToFile("optimize.params.final", values);

    computation.StopComputeNodes();
}

/////////////////////////////////////////////////////////////////
// RunPredictionMode()
//
// Run CONTRAfold in prediction mode.
/////////////////////////////////////////////////////////////////

void RunPredictionMode(const std::vector<std::string> &filenames,
                       const std::string &output_parens_destination,
                       const std::string &output_bpseq_destination,
                       const std::string &output_posteriors_destination)
{
    // retrieve parameter values
    
    Parameters params;
    std::vector<double> values = 
        (params_filename != "") ?
        params.ReadFromFile(params_filename) :
        (toggle_complementary_only ? params.GetDefaultComplementaryValues() : params.GetDefaultValues());
    
    if (gammar < 0)
    {
        // create directories for storing each run
        
        MakeDirectory(output_parens_destination);
        MakeDirectory(output_bpseq_destination);
        MakeDirectory(output_posteriors_destination);
        
        // try different values of gamma
        
        for (int k = -5; k <= 10; k++)
        {
            gammar = Pow(2.0, double(k));
            
            std::string output_parens_subdestination;
            std::string output_bpseq_subdestination;
            std::string output_posteriors_subdestination;
            
            // set output filename or directory for each gamma
            
            if (output_parens_destination != "")
                output_parens_subdestination = SPrintF("%s/%s.gamma=%lf", 
                                                       output_parens_destination.c_str(), 
                                                       output_parens_destination.c_str(), gammar);
            if (output_bpseq_destination != "")
                output_bpseq_subdestination = SPrintF("%s/%s.gamma=%lf", 
                                                      output_bpseq_destination.c_str(), 
                                                      output_bpseq_destination.c_str(), gammar);
            if (output_posteriors_destination != "")
                output_posteriors_subdestination = SPrintF("%s/%s.gamma=%lf", 
                                                           output_posteriors_destination.c_str(), 
                                                           output_posteriors_destination.c_str(), gammar);
            
            // create output directories for output files, if needed
            
            if (filenames.size() > 1)
            {
                MakeDirectory(output_parens_subdestination);
                MakeDirectory(output_bpseq_subdestination);
                MakeDirectory(output_posteriors_subdestination);
            }
            
            // create a "computation" object which will handle all requests for
            // performing heavy-duty computations (e.g., parsing sequences, computing
            // logloss gradients, etc).
            
            Computation computation(filenames,
                                    output_parens_subdestination,
                                    output_bpseq_subdestination,
                                    output_posteriors_subdestination,
                                    toggle_complementary_only,
                                    toggle_use_constraints, 
                                    toggle_partition, 
                                    toggle_viterbi,
                                    toggle_verbose,
                                    posterior_cutoff);
            
            // do predictions using a specified value of gamma
            
            if (computation.IsComputeNode())
            {
                computation.RunComputeNode();
            }
            else
            {
                computation.PredictStructure(computation.GetAllUnits(), values, gammar, false);
                computation.StopComputeNodes();
            }      
        }
        
    }
    else
    {
        
        // create output directories for output files, if needed
        
        if (filenames.size() > 1)
        {
            MakeDirectory(output_parens_destination);
            MakeDirectory(output_bpseq_destination);
            MakeDirectory(output_posteriors_destination);
        }
        
        // create a "computation" object which will handle all requests for
        // performing heavy-duty computations (e.g., parsing sequences, computing
        // logloss gradients, etc).
        
        Computation computation(filenames,
                                output_parens_destination,
                                output_bpseq_destination,
                                output_posteriors_destination,
                                toggle_complementary_only,
                                toggle_use_constraints, 
                                toggle_partition, 
                                toggle_viterbi,
                                toggle_verbose,
                                posterior_cutoff);
        
        // do predictions using a single value of gamma
        
        if (computation.IsComputeNode())
        {
            computation.RunComputeNode();
        }
        else
        {
            computation.PredictStructure(computation.GetAllUnits(), values, gammar, false);
            computation.StopComputeNodes();
        }      
        
    }
}

