/////////////////////////////////////////////////////////////////
// DistributedComputation.hpp
//
// Class for performing distributed optimization.  To use this
// class you must do two things:
//
// (1) Create a subclass of WorkUnit which implements the
//     following methods:
//
//       Execute()
//       GetDescriptionSize()
//       GetEstimatedTime()
//       GetSummary()
//
// (2) Create a DistributedComputation object.  Then, call the
//     RunMasterNode() routine, supplying a vector of pointers 
//     to WorkUnit objects, in order to perform the computation 
//     over all the nodes.
/////////////////////////////////////////////////////////////////

#ifndef DISTRIBUTEDCOMPUTATION_HPP
#define DISTRIBUTEDCOMPUTATION_HPP

#ifdef MULTI
#include <mpi.h>
#endif
#include <cmath>
#include "Utilities.hpp"

/////////////////////////////////////////////////////////////////
// class WorkUnit
//
// Class for storing descriptions of work units to be performed.
/////////////////////////////////////////////////////////////////

class WorkUnit
{
public:
    virtual ~WorkUnit() {}
    virtual int GetDescriptionSize() const = 0;
    virtual double GetEstimatedTime() const = 0;
    virtual std::string GetSummary() const = 0;
};

/////////////////////////////////////////////////////////////////
// WorkUnitComparator()
//
// Comparator to allow sorting of pointers to WorkUnit 
// objects in decreasing order of estimated time.
/////////////////////////////////////////////////////////////////

struct WorkUnitComparator
{
    bool operator()(const WorkUnit *lhs, const WorkUnit *rhs) const;
};

/////////////////////////////////////////////////////////////////
// enum CommandType
/////////////////////////////////////////////////////////////////

enum CommandType
{ 
    CommandType_LoadParameters, 
    CommandType_DoWork, 
    CommandType_GetResultSize,
    CommandType_GetResult, 
    CommandType_ClearResult,
    CommandType_Quit
};

/////////////////////////////////////////////////////////////////
// class DistributedComputation
//
// Wrapper class for performing distributed optimization.
/////////////////////////////////////////////////////////////////

class DistributedComputation
{
    bool toggle_verbose;
    double processing_time;
    double total_time;
    int id;
    int num_procs;
    
public:
    
    DistributedComputation(bool toggle_verbose);
    virtual ~DistributedComputation(){}
    
    void StopComputeNodes();
    
    void RunMasterNode(std::vector<double> &ret, 
                       std::vector<WorkUnit *> descriptions, 
                       std::vector<double> params);
    
    bool IsComputeNode() const;
    bool IsMasterNode() const;
    int GetNumNodes() const;
    int GetNodeID() const;
    void RunComputeNode();
    
    double GetEfficiency() const;
    double GetElapsedTime();
    void ProgressMessage(const std::string &message);
    
    virtual void DoComputation(std::vector<double> &ret, 
                               const WorkUnit *description,
                               const std::vector<double> &params) = 0;
    
};

#endif
