/////////////////////////////////////////////////////////////////
// InferenceEngine.hpp
/////////////////////////////////////////////////////////////////

#ifndef INFERENCEENGINE_HPP
#define INFERENCEENGINE_HPP

#include <queue>
#include <vector>
#include <string>
#include "SStruct.hpp"
#include "Parameters.hpp"
#include "Utilities.hpp"
#include "LogSpace.hpp"

/////////////////////////////////////////////////////////////////
// class InferenceEngine
/////////////////////////////////////////////////////////////////

template<class RealT>
class InferenceEngine {
    
    enum TRACEBACK_TYPE {
#if PARAMS_HELIX_LENGTH || PARAMS_ISOLATED_BASE_PAIR
        TB_FN_HAIRPIN,
        TB_FN_SINGLE,
        TB_FN_BIFURCATION,
        TB_FE_STACKING,
        TB_FE_FN,
        TB_FC_FN,
        TB_FC_HELIX,
        TB_FC_FE,
#else
        TB_FC_HAIRPIN,
        TB_FC_SINGLE,
        TB_FC_BIFURCATION,
#endif
        TB_FM1_PAIRED,
        TB_FM1_UNPAIRED,
        TB_FM_BIFURCATION,
        TB_FM_UNPAIRED,
        TB_FM_FM1,
        TB_F5_ZERO,
        TB_F5_UNPAIRED,
        TB_F5_BIFURCATION,
        NUM_TRACEBACK_TYPES
    };
    
    Parameters params;
    const int L, SIZE;
    bool toggle_complementary_only;
    int is_complementary[M+1][M+1];
    
    int *s, *offset;
    int *allow_unpaired_position;
    int *allow_unpaired, *allow_paired;
    RealT *loss_unpaired_position;
    RealT *loss_unpaired, *loss_paired;
    
    // dynamic programming matrices
    
    int *FCt, *F5t, *FMt, *FM1t;            // traceback
    RealT *FCv, *F5v, *FMv, *FM1v;          // Viterbi  
    RealT *FCi, *F5i, *FMi, *FM1i;          // inside
    RealT *FCo, *F5o, *FMo, *FM1o;          // outside
    
#if PARAMS_HELIX_LENGTH || PARAMS_ISOLATED_BASE_PAIR
    int *FEt, *FNt;
    RealT *FEv, *FNv;
    RealT *FEi, *FNi;
    RealT *FEo, *FNo;
#endif
    
    RealT *posterior;
    
    // cache
    
    RealT cache_score_single[C_MAX_SINGLE_LENGTH+1][C_MAX_SINGLE_LENGTH+1];
    RealT cache_counts_single[C_MAX_SINGLE_LENGTH+1][C_MAX_SINGLE_LENGTH+1];
    RealT *cache_score_helix_sums;
    RealT *cache_counts_helix_sums;
    
    // parameters
    
#if PARAMS_BASE_PAIR
    RealT score_base_pair[M+1][M+1];
#endif
#if PARAMS_BASE_PAIR_DIST
    RealT score_base_pair_dist_at_least[D_MAX_BP_DIST_THRESHOLDS+1];
    RealT cache_score_base_pair_dist[BP_DIST_LAST_THRESHOLD+1];
#endif
#if PARAMS_TERMINAL_MISMATCH
    RealT score_terminal_mismatch[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HAIRPIN_LENGTH
    RealT score_hairpin_length_at_least[D_MAX_HAIRPIN_LENGTH+1];
    RealT cache_score_hairpin_length[D_MAX_HAIRPIN_LENGTH+1];
#endif
#if PARAMS_HAIRPIN_3_NUCLEOTIDES
    RealT score_hairpin_3_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_HAIRPIN_4_NUCLEOTIDES
    RealT score_hairpin_4_nucleotides[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_LENGTH
    RealT score_helix_length_at_least[D_MAX_HELIX_LENGTH+1];
    RealT cache_score_helix_length[D_MAX_HELIX_LENGTH+1];
#endif
#if PARAMS_ISOLATED_BASE_PAIR
    RealT score_isolated_base_pair;
#endif
#if PARAMS_INTERNAL_EXPLICIT
    RealT score_internal_explicit[D_MAX_INTERNAL_EXPLICIT_LENGTH+1][D_MAX_INTERNAL_EXPLICIT_LENGTH+1];
#endif
#if PARAMS_BULGE_LENGTH
    RealT score_bulge_length_at_least[D_MAX_BULGE_LENGTH+1];
#endif
#if PARAMS_INTERNAL_LENGTH
    RealT score_internal_length_at_least[D_MAX_INTERNAL_LENGTH+1];
#endif
#if PARAMS_INTERNAL_SYMMETRY
    RealT score_internal_symmetric_length_at_least[D_MAX_INTERNAL_SYMMETRIC_LENGTH+1];
#endif
#if PARAMS_INTERNAL_ASYMMETRY
    RealT score_internal_asymmetry_at_least[D_MAX_INTERNAL_ASYMMETRY+1];
#endif
#if PARAMS_BULGE_0x1_NUCLEOTIDES
    RealT score_bulge_0x1_nucleotides[M+1];
    RealT score_bulge_1x0_nucleotides[M+1];
#endif
#if PARAMS_BULGE_0x2_NUCLEOTIDES
    RealT score_bulge_0x2_nucleotides[M+1][M+1];
    RealT score_bulge_2x0_nucleotides[M+1][M+1];
#endif
#if PARAMS_BULGE_0x3_NUCLEOTIDES
    RealT score_bulge_0x3_nucleotides[M+1][M+1][M+1];
    RealT score_bulge_3x0_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_INTERNAL_1x1_NUCLEOTIDES
    RealT score_internal_1x1_nucleotides[M+1][M+1];
#endif
#if PARAMS_INTERNAL_1x2_NUCLEOTIDES
    RealT score_internal_1x2_nucleotides[M+1][M+1][M+1];
    RealT score_internal_2x1_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_INTERNAL_2x2_NUCLEOTIDES
    RealT score_internal_2x2_nucleotides[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_STACKING
    RealT score_helix_stacking[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_CLOSING
    RealT score_helix_closing[M+1][M+1];
#endif
#if PARAMS_MULTI_LENGTH
    RealT score_multi_base;
    RealT score_multi_unpaired;
    RealT score_multi_paired;
#endif
#if PARAMS_DANGLE
    RealT score_dangle_left[M+1][M+1][M+1];
    RealT score_dangle_right[M+1][M+1][M+1];
#endif
#if PARAMS_EXTERNAL_LENGTH
    RealT score_external_unpaired;
    RealT score_external_paired;
#endif
    
    // expected counts
    
#if PARAMS_BASE_PAIR
    RealT counts_base_pair[M+1][M+1];
#endif
#if PARAMS_BASE_PAIR_DIST
    RealT counts_base_pair_dist_at_least[D_MAX_BP_DIST_THRESHOLDS];
    RealT cache_counts_base_pair_dist[BP_DIST_LAST_THRESHOLD+1];
#endif
#if PARAMS_TERMINAL_MISMATCH
    RealT counts_terminal_mismatch[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HAIRPIN_LENGTH
    RealT counts_hairpin_length_at_least[D_MAX_HAIRPIN_LENGTH+1];
    RealT cache_counts_hairpin_length[D_MAX_HAIRPIN_LENGTH+1];
#endif
#if PARAMS_HAIRPIN_3_NUCLEOTIDES
    RealT counts_hairpin_3_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_HAIRPIN_4_NUCLEOTIDES
    RealT counts_hairpin_4_nucleotides[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_LENGTH
    RealT counts_helix_length_at_least[D_MAX_HELIX_LENGTH+1];
    RealT cache_counts_helix_length[D_MAX_HELIX_LENGTH+1];
#endif
#if PARAMS_ISOLATED_BASE_PAIR
    RealT counts_isolated_base_pair;
#endif
#if PARAMS_INTERNAL_EXPLICIT
    RealT counts_internal_explicit[D_MAX_INTERNAL_EXPLICIT_LENGTH+1][D_MAX_INTERNAL_EXPLICIT_LENGTH+1];
#endif
#if PARAMS_BULGE_LENGTH
    RealT counts_bulge_length_at_least[D_MAX_BULGE_LENGTH+1];
#endif
#if PARAMS_INTERNAL_LENGTH
    RealT counts_internal_length_at_least[D_MAX_INTERNAL_LENGTH+1];
#endif
#if PARAMS_INTERNAL_SYMMETRY
    RealT counts_internal_symmetric_length_at_least[D_MAX_INTERNAL_SYMMETRIC_LENGTH+1];
#endif
#if PARAMS_INTERNAL_ASYMMETRY
    RealT counts_internal_asymmetry_at_least[D_MAX_INTERNAL_ASYMMETRY+1];
#endif
#if PARAMS_BULGE_0x1_NUCLEOTIDES
    RealT counts_bulge_0x1_nucleotides[M+1];
    RealT counts_bulge_1x0_nucleotides[M+1];
#endif
#if PARAMS_BULGE_0x2_NUCLEOTIDES
    RealT counts_bulge_0x2_nucleotides[M+1][M+1];
    RealT counts_bulge_2x0_nucleotides[M+1][M+1];
#endif
#if PARAMS_BULGE_0x3_NUCLEOTIDES
    RealT counts_bulge_0x3_nucleotides[M+1][M+1][M+1];
    RealT counts_bulge_3x0_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_INTERNAL_1x1_NUCLEOTIDES
    RealT counts_internal_1x1_nucleotides[M+1][M+1];
#endif
#if PARAMS_INTERNAL_1x2_NUCLEOTIDES
    RealT counts_internal_1x2_nucleotides[M+1][M+1][M+1];
    RealT counts_internal_2x1_nucleotides[M+1][M+1][M+1];
#endif
#if PARAMS_INTERNAL_2x2_NUCLEOTIDES
    RealT counts_internal_2x2_nucleotides[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_STACKING
    RealT counts_helix_stacking[M+1][M+1][M+1][M+1];
#endif
#if PARAMS_HELIX_CLOSING
    RealT counts_helix_closing[M+1][M+1];
#endif
#if PARAMS_MULTI_LENGTH
    RealT counts_multi_base;
    RealT counts_multi_unpaired;
    RealT counts_multi_paired;
#endif
#if PARAMS_DANGLE
    RealT counts_dangle_left[M+1][M+1][M+1];
    RealT counts_dangle_right[M+1][M+1][M+1];
#endif
#if PARAMS_EXTERNAL_LENGTH
    RealT counts_external_unpaired;
    RealT counts_external_paired;
#endif
    
    int ComputeRowOffset(int i, int N);
    
    RealT ScoreJunctionA(int i, int j) const;
    RealT ScoreJunctionB(int i, int j) const;
    RealT ScoreBasePair(int i, int j) const;
    RealT ScoreHairpin(int i, int j) const;
    RealT ScoreHelix(int i, int j, int m) const;
    RealT ScoreSingleNucleotides(int i, int j, int p, int q) const;
    RealT ScoreSingle(int i, int j, int p, int q) const;
    
    void CountJunctionA(int i, int j, RealT value);
    void CountJunctionB(int i, int j, RealT value);
    void CountBasePair(int i, int j, RealT value);
    void CountHairpin(int i, int j, RealT value);
    void CountHelix(int i, int j, int m, RealT value);
    void CountSingleNucleotides(int i, int j, int p, int q, RealT value);
    void CountSingle(int i, int j, int p, int q, RealT value);
    
    int EncodeTraceback(int i, int j) const;
    std::pair<int,int> DecodeTraceback(int s) const;
    
    void InitializeScores();
    void InitializeCounts();
    void InitializeCacheCounts();
    
    void BuildCacheScores();
    void ConvertCacheCounts();
    void WriteCounts(std::vector<RealT> &counts);
    
public:
    
    InferenceEngine(const std::string &sequence, bool toggle_complementary_only);
    ~InferenceEngine();		   
    
    void LoadParameters(const std::vector<RealT> &values);
    void UseLoss(const std::vector<int> &true_mapping, const std::vector<RealT> &loss);
    void UseMapping(const std::vector<int> &true_mapping);
    void ComputeViterbi();
    RealT GetViterbiScore() const;
    std::vector<int> PredictPairingsViterbi() const;
    std::vector<RealT> ComputeViterbiFeatureCounts();
    
    void ComputeInside();
    RealT ComputePartitionCoefficient() const;
    void ComputeOutside();
    std::vector<RealT> ComputeFeatureCountExpectations();
    void ComputePosterior();
    std::vector<int> PredictPairingsPosterior(const double gammar) const;
    RealT *GetPosterior(const RealT posterior_cutoff) const;
};

#include "InferenceEngine.ipp"

#endif
