/////////////////////////////////////////////////////////////////
// SStruct.cpp
/////////////////////////////////////////////////////////////////

#include "SStruct.hpp"

/////////////////////////////////////////////////////////////////
// SStruct::IsComplementary()
//
// Check if a base-pairing is one of AU, CG, or GU.
/////////////////////////////////////////////////////////////////

inline bool SStruct::IsComplementary(char c, char d)
{
    c = toupper(c);
    d = toupper(d);
    
    return 
        (c == 'A' && d == 'U') ||
        (c == 'U' && d == 'A') ||
        (c == 'C' && d == 'G') ||
        (c == 'G' && d == 'C') ||
        (c == 'G' && d == 'U') ||
        (c == 'U' && d == 'G');
}

/////////////////////////////////////////////////////////////////
// SStruct::AssertValidSequence()
//
// Check that a string represents a valid sequence.
/////////////////////////////////////////////////////////////////

void SStruct::AssertValidSequence(const std::string &sequence) const
{
    for (size_t i = 1; i < sequence.length(); i++)
    {
        char ch = toupper(sequence[i]);
        if (ch != 'A' && ch != 'C' && ch != 'G' && ch != 'U' && ch != 'N')
        {
            Error("Nucleotide %d has invalid type '%c'.", i, ch);
        }
    }
}

/////////////////////////////////////////////////////////////////
// SStruct::AssertValidMapping()
//
// Check that a std::vector<int> represents a valid secondary
// structure mapping.  Pseudoknots are allowed.
/////////////////////////////////////////////////////////////////

void SStruct::AssertValidMapping(const std::vector<int> &mapping) const
{
    if (mapping.size() == 0) return;
    for (size_t i = 1; i < mapping.size(); i++)
    {
        if (mapping[i] != UNPAIRED && mapping[i] != UNKNOWN)
        {
            if (mapping[i] < 1 || mapping[i] >= int(mapping.size()))
                Error("Position %d of sequence maps to invalid position.", i);
            if (mapping[mapping[i]] != int(i))
                Error("Positions %d and %d of sequence do not map to each other.", i, mapping[i]);
            if (mapping[i] == int(i))
                Error("Position %d of sequence maps to itself.", i);
        }
    }
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from file.  Attempt to detect the format of the
// file automatically.
/////////////////////////////////////////////////////////////////

SStruct::SStruct(const std::string &filename) : 
    sequence("@"), mapping(1, UNKNOWN), name("unnamed")
{
    
    std::ifstream data(filename.c_str());
    if (data.fail()) Error(("Unable to open input file: " + filename).c_str());
    
    // look for first blank line
    
    std::string s;
    while (std::getline(data, s))
    {
        if (s.length() > 0) break;
    }
    
    // analyze to determine file format
    
    FileFormat format;
    if (s[0] == '>')
    {
        format = FileFormat_FASTA;
    }
    else
    {
        std::istringstream iss(s);
        int x, y;
        std::string t;
        if (iss >> x >> t >> y)
        {
            format = FileFormat_BPSEQ;
        }
        else
        {
            format = FileFormat_RAW;
        }
    }
    
    data.close();
    
    // now load file
    
    switch (format)
    {
        case FileFormat_FASTA: LoadFASTA(filename); break;
        case FileFormat_RAW: LoadRAW(filename); break;
        case FileFormat_BPSEQ: LoadBPSEQ(filename); break;
        default: Assert(false, "Should not get here.");
    }
    
    // convert letters
    
    for (size_t i = 1; i < sequence.length(); i++)
    {
        if (sequence[i] == 't') sequence[i] = 'u';
        if (sequence[i] == 'T') sequence[i] = 'U';
        if (!isalpha(sequence[i]))
            Error("Unexpected character '%c' in sequence.", sequence[i]);      
        
        switch (sequence[i])
        {
            case 't': sequence[i] = 'u'; break;
            case 'T': sequence[i] = 'U'; break;
            case 'a': case 'c': case 'g': case 'u': break;
            case 'A': case 'C': case 'G': case 'U': break;
            default:
                if (isupper(sequence[i]))
                {
                    sequence[i] = 'N';
                }
                else
                {
                    sequence[i] = 'n';
                }
        }
    }
    
    AssertValidSequence(sequence);
    AssertValidMapping(mapping);
}
  
/////////////////////////////////////////////////////////////////
// SStruct::~SStruct()
//
// Destructor.
/////////////////////////////////////////////////////////////////

SStruct::~SStruct()
{}

/////////////////////////////////////////////////////////////////
// SStruct::LoadFASTA()
//
// Create object from a FASTA file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadFASTA(const std::string &filename)
{
    std::string s;
    
    std::ifstream data(filename.c_str());
    if (data.fail()) Error(("Unable to open input file: " + filename).c_str());
    
    // retrieve sequence header
    
    while (std::getline(data, s))
    {
        if (s.length() == 0) continue;
        Assert(s[0] == '>', "Auto-detection returned incorrect result.");
        name = s.substr(1);
        if (s.length() > 0) break;
    }
    
    // now retrieve sequence data
    
    while (std::getline(data, s))
    {
        if (s.length() == 0) continue;
        for (size_t i = 0; i < s.length(); i++)
        {
            sequence.push_back (s[i]);
            mapping.push_back (UNKNOWN);
        }
    }  
}

/////////////////////////////////////////////////////////////////
// SStruct::LoadRAW()
//
// Create object from raw (unformatted) file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadRAW(const std::string &filename)
{
    std::string s;
    
    std::ifstream data(filename.c_str());
    if (data.fail()) Error(("Unable to open input file: " + filename).c_str());
    
    // now retrieve sequence data
    
    while (std::getline(data, s))
    {
        if (s.length() == 0) continue;
        for (size_t i = 0; i < s.length(); i++){
            sequence.push_back(s[i]);
            mapping.push_back(UNKNOWN);
        }
    }  
}

/////////////////////////////////////////////////////////////////
// SStruct::LoadBPSEQ()
//
// Create object from BPSEQ file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadBPSEQ(const std::string &filename)
{
    int index, maps_to;
    char ch;
    
    std::ifstream data(filename.c_str());
    if (data.fail()) Error(("Unable to open input file: " + filename).c_str());
    
    while (data >> index >> ch >> maps_to)
    {
        if (index < 1) Error("Only positive sequence indices allowed.");
        if (index != int(mapping.size())) Error("Sequence indices must appear in sorted order.");
        sequence.push_back(ch);
        mapping.push_back(maps_to);
    }
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from given sequence, secondary structure
// mapping, and name.
/////////////////////////////////////////////////////////////////

SStruct::SStruct(const std::string &sequence, 
                 const std::vector<int> &mapping, 
                 const std::string &name) :
    sequence(sequence), mapping(mapping), name (name)
{
    Assert(sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
    Assert(mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
    Assert(sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
    AssertValidSequence(sequence);
    AssertValidMapping(mapping);
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from existing SStruct object.
/////////////////////////////////////////////////////////////////

SStruct::SStruct(const SStruct &rhs) : 
  sequence(rhs.sequence), mapping(rhs.mapping), name (rhs.name)
{
    Assert(sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
    Assert(mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
    Assert(sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
    AssertValidSequence(sequence);
    AssertValidMapping(mapping);
}

/////////////////////////////////////////////////////////////////
// SStruct::operator=
//
// Assignment operator.
/////////////////////////////////////////////////////////////////

const SStruct& SStruct::operator=(const SStruct &rhs)
{
    if (this != &rhs)
    {
        sequence = rhs.sequence;
        mapping = rhs.mapping;
        name = rhs.name;
    }
    
    Assert(sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
    Assert(mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
    Assert(sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
    AssertValidSequence(sequence);
    AssertValidMapping(mapping);
    return *this;
}

/////////////////////////////////////////////////////////////////
// SStruct::GetComplementaryMapping()
//
// Get mapping containing only complementary pairs.
/////////////////////////////////////////////////////////////////

std::vector<int> SStruct::GetComplementaryMapping() const
{
    std::vector<int> new_mapping = mapping;
    
    for (size_t i = 1; i < new_mapping.size(); i++)
    {
        if (new_mapping[i] != UNPAIRED && new_mapping[i] != UNKNOWN && 
            !IsComplementary(sequence[i], sequence[new_mapping[i]]))
        {
            new_mapping[i] = UNPAIRED;
        }
    }
    
    return new_mapping;
}

/////////////////////////////////////////////////////////////////
// SStruct::WriteBPSEQ()
//
// Write sequence in BPSEQ format.
/////////////////////////////////////////////////////////////////

void SStruct::WriteBPSEQ(std::ostream &outfile, bool toggle_complementary_only) const
{
    std::vector<int> new_mapping = toggle_complementary_only ? GetComplementaryMapping() : mapping;
    
    for (size_t i = 1; i < new_mapping.size(); i++)
        outfile << i << ' ' << sequence[i] << ' ' << new_mapping[i] << std::endl;
}

/////////////////////////////////////////////////////////////////
// SStruct::WriteParens()
//
// Write sequence in parenthesized format.  This routine assumes
// that the structure does not contain pseudoknots.
/////////////////////////////////////////////////////////////////

void SStruct::WriteParens(std::ostream &outfile, bool toggle_complementary_only) const
{
    std::vector<int> new_mapping = toggle_complementary_only ? GetComplementaryMapping() : mapping;
    
    outfile << '>' << name << std::endl;
    for (size_t i = 1; i < sequence.length(); i++)
        outfile << sequence[i];
    outfile << std::endl;
    for (size_t i = 1; i < new_mapping.size(); i++)
    {
        if (new_mapping[i] == UNKNOWN)
        {
            outfile << '?';
        }
        else if (new_mapping[i] == UNPAIRED)
        {
            outfile << '.';
        }
        else if (new_mapping[i] > int(i))
        {
            outfile << '(';
        }
        else
        {
            outfile << ')';
        }
    }
    outfile << std::endl;
}
